import cv2
from pydantic import BaseModel


class OpenCVCameraOptions(BaseModel):
    port_number: int
    exposure: int = -5
    resolution_width: int = 1280
    resolution_height: int = 720


# # OpenCV Implementation of interacting with a camera
# class OpenCVCamera:
#
#     def __init__(self, options: OpenCVCameraOptions):
#         self._options = options
#
#     def connect(self):
#         # c = cv2.VideoCapture()
#         cv2.VideoCapture(, cv2.CAP_ANY)



import bpy
import numpy as np
from mathutils import Matrix, Vector, Euler
import pathlib 

class OMC_OT_loadOpenMoCapSession(bpy.types.Operator): #setting the type as "OMC" for "OpenMoCap"
    """ Load data from an OpenMoCap(OMC) recording session into Blender """
    bl_idname = "omc.load_session_data"
    bl_label = "Load Session Data"
    bl_options = {'REGISTER', 'UNDO'}

    charucoMar_size: bpy.props.FloatProperty(
        name = 'Charuco Marker Size',
        description="Radius of the Charuco Marker Spheres",
        default = .2,
        min = 0, soft_max = 1,
    )
    
    dlcMar_size: bpy.props.FloatProperty(
        name = 'DLC Marker Size',
        description="Radius of the DLC Marker Spheres",
        default = .3,
        min = 0, soft_max = 1,
    )

    bodyMar_size: bpy.props.FloatProperty(
        name = 'Skeleton Body Marker Size',
        description="Radius of the Skeleton Body Marker Spheres",
        default = .2,
        min = 0, soft_max = 1,
    )

    faceMar_size: bpy.props.FloatProperty(
        name = 'Skeleton Face Marker Size',
        description="Radius of the Skeleton Face Marker Spheres",
        default = .15,
        min = 0, soft_max = 1,
    )

    handMar_size: bpy.props.FloatProperty(
        name = 'Skeleton Hand Marker Size',
        description="Radius of the Skeleton Hand Marker Spheres",
        default = .2,
        min = 0, soft_max = 1,
    )

    meshType: bpy.props.StringProperty(
        name = 'Marker Mesh Type',
        description = 'Type of mesh to draw at marker locations (just `sphere` and `monkey` for now)',
        default = 'uv_sphere',
    )

    def execute(self, context):
        """Load in data from an OpenMoCap recording session and place all 3D points as spheres in the 3D Viewport"""
        C = bpy.context
        D = bpy.data

        # #%% Nuke it all 
        # bpy.ops.object.select_all(action='SELECT') #select all
        # bpy.ops.object.delete(use_global=False) #delete everything 

        # try:
        #     bpy.ops.outliner.select_all(action='SELECT')
        #     bpy.ops.outliner.delete(hierarchy=True)
        # except:
        #     pass

        #%% _______________________________________________________________________
        #  Load in data

        loadedData  = np.load(r"C:\Users\jonma\Dropbox\GitKrakenRepos\OpenMoCap\Data\test6_01_21a\outData\saveData.npz")
        skel_fr_mar_xyz = loadedData['skel_fr_mar_xyz'] #OpenPose skeleton data
        dlcPts_fr_mar_xyz = loadedData['dlcPts_fr_mar_xyz']
        charucoCorners_pt_XYZ = loadedData['charucoCornersXYZ']

        startFr = 0
        numFrames = skel_fr_mar_xyz.shape[0]-1

        #names of markers 
        skel_markerID = ["Nose", "Neck", "RShoulder", "RElbow", "RWrist", "LShoulder",
        "LElbow", "LWrist", "MidHip", "RHip", "RKnee", "RAnkle", "LHip", "LKnee",
        "LAnkle", "REye", "LEye", "REar", "LEar", "LBigToe", "LSmallToe", "LHeel",
        "RBigToe", "RSmallToe", "RHeel"]

        #%% create spheres at marker locations

        #%% _______________________________________________________________________
        #Charuco grid!
        charucoCollectionName = 'CharucoBoard'
        # charColl = eby.create_collection(charucoCollectionName)
        # assert charColl.name == charucoCollectionName, 'Charuco Collection name doesnt match. Did it already exist when you tried to make it? If so, Blender would make a new one called `thing.001` or something' 

        # C.scene.collection.children.link(charColl) #link the newly created collection to the Blender Scene (so it will show up in the outliner)

        print('Loading Charuco Markers')
        for marNum in range(len(charucoCorners_pt_XYZ[:,0])):
            thisMarLoc = charucoCorners_pt_XYZ[marNum,:]
            ms =   self.charucoMar_size

            if self.meshType == 'uv_sphere':
                bpy.ops.mesh.primitive_uv_sphere_add(align='WORLD', location=thisMarLoc, scale=(ms, ms, ms))            
            elif self.meshType == 'monkey':
                bpy.ops.mesh.primitive_monkey_add(align='WORLD', location=thisMarLoc, scale=(ms, ms, ms))
                
            thisMarker = C.active_object

            # eby.move_object_to_collection(thisMarker, charColl)

        #%% _______________________________________________________________________
        # Deeplabcut (DLC) data!
        dlcCollectionName = 'DLC'
        # dlcColl = eby.create_collection(dlcCollectionName)
        # assert dlcColl.name == dlcCollectionName, 'Charuco Collection name doesnt match. Did it already exist when you tried to make it? If so, Blender would make a new one called `thing.001` or something' 

        # C.scene.collection.children.link(charColl) #link the newly created collection to the Blender Scene (so it will show up in the outliner)
        
        print('Loading DLC Markers')
        for marNum in range(len(dlcPts_fr_mar_xyz[0,:,0])):
            thisMarLoc = dlcPts_fr_mar_xyz[0,marNum,:]
            ms = self.dlcMar_size
            
            if self.meshType == 'uv_sphere':
                bpy.ops.mesh.primitive_uv_sphere_add(align='WORLD', location=thisMarLoc, scale=(ms, ms, ms))            
            elif self.meshType == 'monkey':
                bpy.ops.mesh.primitive_monkey_add(align='WORLD', location=thisMarLoc, scale=(ms, ms, ms))
            
            thisMarker = C.active_object
            # eby.move_object_to_collection(thisMarker, dlcColl)

            #loop through each frame (after the first [0th] frame) to set the keyframes for this marker
            for fr in range(1, numFrames):
                thisMarker.location = dlcPts_fr_mar_xyz[fr,marNum,:]
                thisMarker.keyframe_insert(data_path="location", frame=fr)



        #%% _______________________________________________________________________
        # Skreleton data!
        skelCollectionName = 'skel'
        # skelColl = eby.create_collection(skelCollectionName)
        # assert skelColl.name == skelCollectionName, 'Charuco Collection name doesnt match. Did it already exist when you tried to make it? If so, Blender would make a new one called `thing.001` or something' 

        # C.scene.collection.children.link(charColl) #link the newly created collection to the Blender Scene (so it will show up in the outliner)
        print('Loading Skeleton Markers!')
        for marNum in range(len(skel_fr_mar_xyz[startFr,:,0])):
            thisMarLoc = skel_fr_mar_xyz[startFr,marNum,:]
            
            #these will define the size of teh body, hand, and face markers
            bms = self.bodyMar_size
            hms = self.handMar_size
            fms = self.faceMar_size

            if self.meshType == 'uv_sphere':
                bpy.ops.object.armature_add(align='WORLD', location=thisMarLoc)            
            elif self.meshType == 'monkey':
                bpy.ops.mesh.primitive_monkey_add(align='WORLD', location=thisMarLoc)
            
            thisMarker = C.active_object

            # eby.move_object_to_collection(thisMarker, skelColl)
            
            #get names of body markers from name array "skel_markerID" (and build hand and face names from there)
            if marNum < 25 :
                thisMarker.name = skel_markerID[marNum]
                thisMarker.scale = (bms, bms, bms)
            elif marNum < 46:                
                if marNum == 26: print('Loading Hands')                
                thisMarker.name = "HandR"
                thisMarker.scale=(hms, hms, hms)
            elif marNum < 67:
                thisMarker.name = "HandL" 
                thisMarker.scale=(hms, hms, hms)
            else:
                if marNum == 67: print('Loading Face')
                thisMarker.name = 'Face'
                thisMarker.scale=(fms, fms, fms)

            #loop through each frame (after the first [0th] frame) to set the keyframes for this marker
            for fr in range(1, numFrames):
                thisMarker.location = skel_fr_mar_xyz[fr,marNum,:]
                thisMarker.keyframe_insert(data_path="location", frame=fr)

        return{'FINISHED'}    




##
##
##  ███████ ███    ███  ██████     ███████ ███████ ███████ ███████ ██  ██████  ███    ██ 
##  ██      ████  ████ ██          ██      ██      ██      ██      ██ ██    ██ ████   ██ 
##  █████   ██ ████ ██ ██          ███████ █████   ███████ ███████ ██ ██    ██ ██ ██  ██ 
##  ██      ██  ██  ██ ██               ██ ██           ██      ██ ██ ██    ██ ██  ██ ██ 
##  ██      ██      ██  ██████     ███████ ███████ ███████ ███████ ██  ██████  ██   ████ 
##                                                                                       
## https://patorjk.com/software/taag/#p=display&f=ANSI%20Regular&t=FMC%20Session

from FMC_MultiCamera import FMC_MultiCamera
from FMC_Visualizer import FMC_Visualizer

import datetime
from pathlib import Path
import pickle 

from rich import print, inspect
from rich.console import Console
import numpy as np 

from freemocap import fmc_anipose
from aniposelib.boards import CharucoBoard

rich_console = Console()


class FMC_Session:
    ## 
    ## 
    ##                 ██ ███    ██ ██ ████████                 
    ##                 ██ ████   ██ ██    ██                    
    ##                 ██ ██ ██  ██ ██    ██                    
    ##                 ██ ██  ██ ██ ██    ██                    
    ## ███████ ███████ ██ ██   ████ ██    ██    ███████ ███████ 
    ##                                                          
    ## 

    def __init__(self, sessionID=None,  freemocap_data_folder=None, rotation_codes_list=None):
        
        if freemocap_data_path is None:
            user_home_directory = Path.home()
            self.freemocap_data_path = user_home_directory /'FreeMoCap_Data'
            
            if self.freemocap_data_path.exists():
                rich_console.log("FreeMoCap_Data folder found in User home directory - " + str(self.freemocap_data_path))
            elif not self.freemocap_data_path.exists():
                self.freemocap_data_path.mkdir(exist_ok=True)
                rich_console.log("FreeMoCap_Data CREATED in  User home directory - " + str(self.freemocap_data_path))
                rich_console.rule("To create a FreeMoCap_Data folder in a different place, use `FMC_Session(freemocap_data_folder=\"path_to_folder\")`")
        else:
            self.freemocap_data_path = freemocap_data_path
           
        if sessionID is None:
            self.sessionID = datetime.datetime.now().strftime("FreeMoCap_Session_%Y-%m-%d_%H_%M_%S")
        else:
            self.sessionID = sessionID

    ##   
    ##   ██████  ███████  ██████     ███    ███ ██    ██ ██      ████████ ██      ██████  █████  ███    ███ 
    ##   ██   ██ ██      ██          ████  ████ ██    ██ ██         ██    ██     ██      ██   ██ ████  ████ 
    ##   ██████  █████   ██          ██ ████ ██ ██    ██ ██         ██    ██     ██      ███████ ██ ████ ██ 
    ##   ██   ██ ██      ██          ██  ██  ██ ██    ██ ██         ██    ██     ██      ██   ██ ██  ██  ██ 
    ##   ██   ██ ███████  ██████     ██      ██  ██████  ███████    ██    ██      ██████ ██   ██ ██      ██ 
    ##                                                                                                      
    ##                                                                                                      
            
    def record_multi_cam(self,freemocap_data_folder=None, rotation_codes_list=None):
        self.multi_cam = FMC_MultiCamera(freemocap_data_folder=str(freemocap_data_path), rotation_codes_list=in_rotation_codes_list, rec_name=self.sessionID )
        self.multi_cam.start()    
    
    ##          
    ##          
    ##   ██████  █████  ██      ██ ██████       ██████  █████  ██████      ██    ██  ██████  ██      
    ##  ██      ██   ██ ██      ██ ██   ██     ██      ██   ██ ██   ██     ██    ██ ██    ██ ██      
    ##  ██      ███████ ██      ██ ██████      ██      ███████ ██████      ██    ██ ██    ██ ██      
    ##  ██      ██   ██ ██      ██ ██   ██     ██      ██   ██ ██           ██  ██  ██    ██ ██      
    ##   ██████ ██   ██ ███████ ██ ██████       ██████ ██   ██ ██            ████    ██████  ███████ 
    ##                                                                                               
    ##  
    
    def camera_capture_volume_calibration(self, calibration_method="anipose_charuco", charucoSquareSize=36):
        
        num_charuco_rows = 7
        num_charuco_cols = 5
        num_charuco_tracked_points = (num_charuco_rows-1) * (num_charuco_cols-1)
        board = CharucoBoard(num_charuco_rows, num_charuco_cols,
                    #square_length=1, # here, in mm but any unit works (JSM NOTE - just using '1' so resulting units will be in 'charuco squarelenghts`)
                    #marker_length=.8,
                    #  square_length = 121, #big boi charuco
                    #  marker_length = 98,
                    square_length = charucoSquareSize,#mm
                    marker_length = charucoSquareSize*.8,#mm
                    marker_bits=4, dict_size=250)
        
        
        
        self.synched_video_path_list = self.multi_cam._each_cam_video_filename_list
        
        video_paths_list_of_list_of_strings = []
        
        for this_path in self.synched_video_path_list:
            video_paths_list_of_list_of_strings.append([str(this_path)])#anipose needs this to be a list of lists (which is annoying by whatevs)

        cam_names = ['cam_'+str(cam_num) for cam_num in range(len(self.synched_video_path_list))]
        
        self.anipose_camera_calibration = fmc_anipose.CameraGroup.from_names(cam_names, fisheye=True  )  # Looking through their code... it looks lke the 'fisheye=True' doesn't do much (see 2020-03-29 obsidian note)    
        error,charuco_frame_data, charuco_frame_numbers = self.anipose_camera_calibration.calibrate_videos(video_paths_list_of_list_of_strings, board)   

        print("Anipose Calibration Successful!")

        #save calibration info to files 
        calibration_toml_filename = "{}_camera_calibration.toml".format(self.multi_cam._rec_name)        
        self.camera_calibration_toml_path = self.multi_cam._save_path / calibration_toml_filename        
        self.anipose_camera_calibration.dump(self.camera_calibration_toml_path) 

        camera_calibration_info_dict = self.anipose_camera_calibration.get_dicts()
        camera_calibration_pickle_path = self.multi_cam._save_path / "{}_camera_calibration.pickle".format(self.multi_cam._rec_name)

        with open(str(camera_calibration_pickle_path), 'wb') as pickle_file:
            pickle.dump(camera_calibration_info_dict, pickle_file)  
            
        #convert charuco data into a format that can be 3d reconstructed (effectively providing dummy data for the rest of the 3d reconstruction pipeline)
        self.charuco_nCams_nFrames_nImgPts_XY = np.empty([self.multi_cam.num_cams, self.multi_cam.num_frames, num_charuco_tracked_points,  2])
        self.charuco_nCams_nFrames_nImgPts_XY[:] = np.nan

        
        for this_charuco_frame_data, this_charuco_frame_num in zip(charuco_frame_data,charuco_frame_numbers):
            for this_cam_num in range(self.multi_cam.num_cams):
                try:
                    self.charuco_nCams_nFrames_nImgPts_XY[this_cam_num, this_charuco_frame_num, :,:] = np.squeeze(this_charuco_frame_data[this_cam_num]["filled"])
                except:
                    # print("failed frame:", frame)
                    continue
    ####            
    ####    
    ####    ██████  ███████  ██████  ██████  ███    ██ ███████ ████████ ██████  ██    ██  ██████ ████████     ██████  ██████  
    ####    ██   ██ ██      ██      ██    ██ ████   ██ ██         ██    ██   ██ ██    ██ ██         ██             ██ ██   ██ 
    ####    ██████  █████   ██      ██    ██ ██ ██  ██ ███████    ██    ██████  ██    ██ ██         ██         █████  ██   ██ 
    ####    ██   ██ ██      ██      ██    ██ ██  ██ ██      ██    ██    ██   ██ ██    ██ ██         ██             ██ ██   ██ 
    ####    ██   ██ ███████  ██████  ██████  ██   ████ ███████    ██    ██   ██  ██████   ██████    ██        ██████  ██████  
    ####                                                                                                                      
    ####                                                                                                                      

    def reconstruct3D(self, data_nCams_nFrames_nImgPts_XYC, calibration_toml_path = None, confidence_threshold=0.3):
                
        if calibration_toml_path: #if the user specified a calibration toml (from a previous Anipose-based calibration) - load it here. Otherwise, assume the calibration is included in `self`                        
            self.anipose_camera_calibration = fmc_anipose.CameraGroup.load(str(calibration_toml_path))

        num_cams, num_frames, num_img_points, num_dims = data_nCams_nFrames_nImgPts_XYC.shape

        if num_dims == 3: #if there are 3 dimensions, the third one is 'confidence' and can be used to threshold out low-confidence tracks
            dataOG = data_nCams_nFrames_nImgPts_XYC.copy()

            for camNum in range(num_cams):
                            
                thisCamX = data_nCams_nFrames_nImgPts_XYC[camNum, :, :,0 ]
                thisCamY = data_nCams_nFrames_nImgPts_XYC[camNum, :, :,1 ]
                thisCamConf = data_nCams_nFrames_nImgPts_XYC[camNum, :, :, 2]

                thisCamX[thisCamConf < confidence_threshold] = np.nan
                thisCamY[thisCamConf < confidence_threshold] = np.nan                                        


        if num_dims == 2:
            data_nCams_nFrames_nImgPts_XY = data_nCams_nFrames_nImgPts_XYC
        elif num_dims == 3:
            data_nCams_nFrames_nImgPts_XY = np.squeeze(data_nCams_nFrames_nImgPts_XYC[:, :, :, 0:2])

        dataFlat_nCams_nTotalPoints_XY = data_nCams_nFrames_nImgPts_XY.reshape(num_cams, -1, 2)  # reshape data to collapse across 'frames' so it becomes [numCams, numFrames*numPoints, XY]

        console.print('Reconstructing 3d points...')
        data3d_flat = self.anipose_camera_calibration.triangulate(dataFlat_nCams_nTotalPoints_XY, progress=True)

        dataReprojerr_flat = self.anipose_camera_calibration.reprojection_error( data3d_flat, dataFlat_nCams_nTotalPoints_XY, mean=True)

        ##return:
        data_fr_mar_xyz = data3d_flat.reshape(num_frames, num_img_points, 3)
        dataReprojErr = dataReprojerr_flat.reshape(num_frames, num_img_points)

        return data_fr_mar_xyz, dataReprojErr
    
    ###    
    ###    
    ###   ████████ ██████   █████   ██████ ██   ██ ██ ███    ██  ██████  
    ###      ██    ██   ██ ██   ██ ██      ██  ██  ██ ████   ██ ██       
    ###      ██    ██████  ███████ ██      █████   ██ ██ ██  ██ ██   ███ 
    ###      ██    ██   ██ ██   ██ ██      ██  ██  ██ ██  ██ ██ ██    ██ 
    ###      ██    ██   ██ ██   ██  ██████ ██   ██ ██ ██   ████  ██████  
    ###                                                                                       
    ###                                                                                       
    def run_tracking_methods(self, use_media_pipe_bool=True, reconstruct3D_bool = True, save_to_npy_bool=True):

        if use_media_pipe_bool:
            from tracking_methods.fmc_mediapipe_alpha import run_mediapipe
            
            self.mediapipe_nCams_nFrames_nImgPts_XY = run_mediapipe(self.synched_video_path_list)
            
            
            if reconstruct3D_bool:
                self.mediapipe_nFrames_nTrackedPoints_XYZ, self.mediapipe_reprojection_error = self.reconstruct3D(self.mediapipe_nCams_nFrames_nImgPts_XY)
                
                            
            if save_to_npy_bool:
                mediapipe3d_filename = self.multi_cam._save_path/'mediapipe_3d_points.npy'
                np.save(mediapipe3d_filename,self.mediapipe_nFrames_nTrackedPoints_XYZ)
                
        f=9

    



    ###         
    ### ███████ ████████  ██████              
    ### ██         ██    ██                   
    ### █████      ██    ██                   
    ### ██         ██    ██                   
    ### ███████    ██     ██████     ██ ██ ██ 
    ###                                       
    
    def __enter__(self):
        """Context manager -  No need to do anything special on start"""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Context manager -  no need to do anything special on stop (maybe set 'exit_event?" """
        return self

###   
###               
###   ██ ███████                     ███    ███  █████  ██ ███    ██                 
###   ██ ██                          ████  ████ ██   ██ ██ ████   ██                 
###   ██ █████                       ██ ████ ██ ███████ ██ ██ ██  ██                 
###   ██ ██                          ██  ██  ██ ██   ██ ██ ██  ██ ██                 
###   ██ ██          ███████ ███████ ██      ██ ██   ██ ██ ██   ████ ███████ ███████ 
###                                                                                  
###                                                                                  

if __name__ == "__main__":
    
    import socket
    this_computer_name = socket.gethostname()
    
    console = Console()
    
    freemocap_data_path=None
    in_rotation_codes_list=None
    charucoSquareSize = 36
    
    if this_computer_name=='jon-hallway-XPS-8930':
        freemocap_data_path = Path('/home/jon/Dropbox/FreeMoCapProject/FreeMocap_Data')
        in_rotation_codes_list = ['cv2.ROTATE_90_COUNTERCLOCKWISE', 'cv2.ROTATE_90_COUNTERCLOCKWISE', 'cv2.ROTATE_90_CLOCKWISE', 'cv2.ROTATE_90_CLOCKWISE', 'cv2.ROTATE_90_CLOCKWISE', ]
    elif this_computer_name == 'DESKTOP-DCG6K4F':
        freemocap_data_path = Path(r'C:\Users\jonma\Dropbox\FreeMoCapProject\FreeMocap_Data')
        this_charucoSquareSize = 123
    
    try:
        sesh =  FMC_Session(sessionID = None, freemocap_data_folder=str(freemocap_data_path), rotation_codes_list=in_rotation_codes_list)
        sesh.record_multi_cam()
        sesh.camera_capture_volume_calibration(charucoSquareSize = this_charucoSquareSize)
        sesh.charuco_nFrames_nTrackedPoints_XYZ, sesh.charuco_reprojection_error = sesh.reconstruct3D(sesh.charuco_nCams_nFrames_nImgPts_XY)
        
        charuco3d_filename = sesh.multi_cam._save_path/'charuco_3d_points.npy'
        np.save(charuco3d_filename,sesh.charuco_nFrames_nTrackedPoints_XYZ)
        
        sesh.run_tracking_methods()
        visualizer = FMC_Visualizer(fmc_session_obj=sesh)
        visualizer.start()
        
    except:
        console.print_exception()


#%% Working from this youtube vid - https://www.youtube.com/watch?v=JrGFQp9njas
from rich import print
from rich import pretty
pretty.install() #makes all print statement output pretty

from rich import inspect

#%%
print('[red]hello[/red] :poop:')
# %%
locals()
#%%
print(locals())

# %%
inspect('asdf')
# %%

inspect('asfdf',methods=True)


# %%
from rich.console import Console
console = Console()  
console.print('[green]hello[/green] :poop:') 
console.log('hello :poop:', style='bold blue') 
console.info('hello :poop:', style='bold green') 
# inspect(console, methods=True)

# %%
from rich.table import Table 

table = Table(show_header=True, header_style="bold magenta")
table.add_column("Date", style="dim", width=12)
table.add_column("Title")
table.add_column("Production Budget", justify="right")
table.add_column("Box Office", justify="right")
table.add_row(
    "Dev 20, 2019", "Star Wars: The Rise of Skywalker", "$275,000,000", "$375,126,118"
)
table.add_row(
    "May 25, 2018",
    "[red]Solo[/red]: A Star Wars Story",
    "$275,000,000",
    "$393,151,347",
)
table.add_row(
    "Dec 15, 2017",
    "Star Wars Ep. VIII: The Last Jedi",
    "$262,000,000",
    "[bold]$1,332,539,889[/bold]",
)

console.print(table)
# %%
from time import sleep
from rich.progress import track 

for step in track(range(10)):
    sleep(0.5)
    step
# %%
from rich.traceback import install as rich_traceback_install
rich_traceback_install()
1/0
# %%



from src.api.routes.camera.camera_route import camera_router
from src.api.routes.health.health_check_route import healthcheck_router

# REGISTER NEW ROUTES HERE
enabled_routers = [
    healthcheck_router,
    camera_router
]



import bpy 
import numpy as np 
from pathlib import Path

class FMC_OT_loadMarkerEmpties(bpy.types.Operator): #setting the type as "FMC" for "OpenMoCap"
    """ Load marker data via FMC yaml config and place empties in Blender scene """
    bl_idname = "fmc.load_marker_empties"
    bl_label = "Load Marker Empties"
    bl_options = {'REGISTER', 'UNDO'}

    emptyType: bpy.props.StringProperty(
        name = 'Empty Type',
        description = 'Type of empty to draw at marker locations (just `sphere` for now)',
        default = 'SPHERE',
    )

    emptySize: bpy.props.FloatProperty(
        name = 'Empty size',
        description = 'Size of the empty markers',
        default = .01,
        min = 0, soft_max =1,
    )

    def execute(self, context):
        configPath = Path(context.scene.fmc_session_path)        
        sessionPath = configPath.parent
        
        try:
            openpose_3d_data_path = sessionPath / 'DataArrays' / 'openPoseSkel_3d.npy'
            openpose_skel_fr_mar_xyz = np.load(str(openpose_3d_data_path))
            #convert to meters
            openpose_skel_fr_mar_xyz = openpose_skel_fr_mar_xyz/1000
            #names of markers 
            openpose_skel_marker_names = ["Nose", "Neck", "RShoulder", "RElbow", "RWrist", "LShoulder",
            "LElbow", "LWrist", "MidHip", "RHip", "RKnee", "RAnkle", "LHip", "LKnee",
            "LAnkle", "REye", "LEye", "REar", "LEar", "LBigToe", "LSmallToe", "LHeel",
            "RBigToe", "RSmallToe", "RHeel"]
            openpose_bool = True
        except:
            openpose_bool = False
            print('No OpenPose Data Found')

        try:
            mediapipe_3d_data_path = sessionPath / 'DataArrays' / 'mediapipeSkel_3d.npy'
            mediapipe_skel_fr_mar_xyz = np.load(str(mediapipe_3d_data_path))
            #convert to meters
            mediapipe_skel_fr_mar_xyz = mediapipe_skel_fr_mar_xyz/1000
            #names of markers 
            mediapipe_skel_marker_names = [ "nose",
                                           
                                            "left_eye_inner",
                                            "left_eye",
                                            "left_eye_outer",
                                           
                                            "right_eye_inner",
                                            "right_eye",
                                            "right_eye_outer",
                                           
                                            "left_ear",
                                            "right_ear",
                                           
                                            "mouth_left",
                                            "mouth_right",
                                           
                                            "left_shoulder",
                                            "right_shoulder",
                                           
                                            "left_elbow",
                                            "right_elbow",
                                           
                                            "left_wrist",
                                            "right_wrist",
                                           
                                            "left_pinky",
                                            "right_pinky",
                                            
                                            "left_index",
                                            "right_index",
                                            
                                            "left_thumb",
                                            "right_thumb",
                                            
                                            "left_hip",
                                            "right_hip",
                                            
                                            "left_knee",
                                            "right_knee",
                                            
                                            "left_ankle",
                                            "right_ankle",
                                            
                                            "left_heel",
                                            "right_heel",
                                            
                                            "left_foot_index",
                                            "right_foot_index",   
                                            ]
            mediapipe_hand_marker_names = ["wrist",
                                           "thumb_cmc",
                                           "thumb_mcp",
                                           "thumb_ip",
                                           "thumb_tip",
                                           
                                           "index_finger_cmc",
                                           "index_finger_mcp",
                                           "index_finger_ip",
                                           "index_finger_tip",
                                           
                                           "middle_finger_cmc",
                                           "middle_finger_mcp",
                                           "middle_finger_ip",
                                           "middle_finger_tip",
                                                                                      
                                           "ring_finger_cmc",
                                           "ring_finger_mcp",
                                           "ring_finger_ip",
                                           "ring_finger_tip",
                                           
                                           "pinky_cmc",
                                           "pinky_mcp",
                                           "pinky_ip",
                                           "pinky_tip",
                                           
                                           
                                            ]
            mediapipe_bool = True
        except:
            mediapipe_bool = False
            print('No mediapipe Data Found')

        
        startFr = 0
        numFrames = mediapipe_skel_fr_mar_xyz.shape[0]-1
        
        bpy.context.scene.frame_end = numFrames
        
        #%% _______________________________________________________________________
        # Skreleton data!
        print('Loading Skeleton Markers!')
        face_iter = 0

        for marNum in range(len(mediapipe_skel_fr_mar_xyz[startFr,:,0])):
            
            
            
            thisMarLoc = mediapipe_skel_fr_mar_xyz[startFr,marNum,:]
            
            #these will define the size of teh body, hand, and face markers
            bms = self.emptySize
            hms = bms *.5
            fms = bms *.5

            emptyType = self.emptyType

            bpy.ops.object.empty_add(type=self.emptyType, align='WORLD', location=thisMarLoc, scale=(bms, bms, bms))
            thisMarker = context.active_object

            #get names of body markers from name array "skel_markerID" (and build hand and face names from there)
            if marNum < len(mediapipe_skel_marker_names) :
                thisMarker.name = mediapipe_skel_marker_names[marNum]
                thisMarker.scale = (bms, bms, bms)
            elif marNum < len(mediapipe_skel_marker_names) + 2*len(mediapipe_hand_marker_names):            
                
                if marNum <  len(mediapipe_skel_marker_names) +len(mediapipe_hand_marker_names):         
                    this_hand_prefix = 'right_hand_'
                    this_hand_prefix = 'right_hand_'
                    marker_num_offset = len(mediapipe_skel_marker_names)
                else:
                    this_hand_prefix = 'left_hand_'
                    marker_num_offset = len(mediapipe_skel_marker_names) + len(mediapipe_hand_marker_names)
                    
                try:
                    print('this hand marnum is ' + str(marNum-marker_num_offset))
                    thisMarker.name = this_hand_prefix + mediapipe_hand_marker_names[marNum-marker_num_offset]
                except:
                    print('something weired in hand town for marker num' + str(marNum))
                    f=9
                    
                thisMarker.scale=(hms, hms, hms)

            else:                
                thisMarker.name = 'face_'+str(face_iter).zfill(4)
                face_iter += 1
                thisMarker.scale=(fms, fms, fms)
  
            #loop through each frame (after the first [0th] frame) to set the keyframes for this marker
            for fr in range(1, numFrames):
                thisMarker.location = mediapipe_skel_fr_mar_xyz[fr,marNum,:]
                thisMarker.keyframe_insert(data_path="location", frame=fr)
            print('loaded marker: '+ thisMarker.name)
            
        print('Done!')    
        return{'FINISHED'}        




"""
============
██████  ██       █████  ██    ██     ███████ ██   ██ ███████ ██      ███████ ████████  ██████  ███    ██      █████  ███    ██ ██ ███    ███  █████  ████████ ██  ██████  ███    ██ 
██   ██ ██      ██   ██  ██  ██      ██      ██  ██  ██      ██      ██         ██    ██    ██ ████   ██     ██   ██ ████   ██ ██ ████  ████ ██   ██    ██    ██ ██    ██ ████   ██ 
██████  ██      ███████   ████       ███████ █████   █████   ██      █████      ██    ██    ██ ██ ██  ██     ███████ ██ ██  ██ ██ ██ ████ ██ ███████    ██    ██ ██    ██ ██ ██  ██ 
██      ██      ██   ██    ██             ██ ██  ██  ██      ██      ██         ██    ██    ██ ██  ██ ██     ██   ██ ██  ██ ██ ██ ██  ██  ██ ██   ██    ██    ██ ██    ██ ██  ██ ██ 
██      ███████ ██   ██    ██        ███████ ██   ██ ███████ ███████ ███████    ██     ██████  ██   ████     ██   ██ ██   ████ ██ ██      ██ ██   ██    ██    ██  ██████  ██   ████                                                                                                                                                                                                                                                                                                                                                                         
============
Font - ANSI Regular - https://patorjk.com/software/taag/#p=display&f=ANSI%20Regular&t=Play%20Skeleton%20Animation
============
Originally based on tutorial from  - https://matplotlib.org/2.1.2/gallery/animation/simple_3danim.html

Create a matplotlib animation from a FreeMoCap recording session.
"""


################################################################################################################
################################################################################################################
###
### ██ ███    ███ ██████   ██████  ██████  ████████ 
### ██ ████  ████ ██   ██ ██    ██ ██   ██    ██    
### ██ ██ ████ ██ ██████  ██    ██ ██████     ██    
### ██ ██  ██  ██ ██      ██    ██ ██   ██    ██    
### ██ ██      ██ ██       ██████  ██   ██    ██    
###
################################################################################################################
################################################################################################################                                                        
                                                        

import numpy as np

import matplotlib as mpl
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3
import matplotlib.animation as animation
from matplotlib.widgets import Slider
import cv2 
from scipy.signal import savgol_filter
import moviepy.editor as mp

import os
import copy
from pathlib import Path

#RICH CONSOLE STUFF
from rich import pretty
pretty.install() #makes all print statement output pretty
from rich import inspect
from rich.console import Console
console = Console()  
from rich.traceback import install as rich_traceback_install
from rich.markdown import Markdown

#colors from Taylor Davis branding - 
humon_dark = np.array([37, 67, 66])/255
humon_green = np.array([53, 93, 95])/255
humon_blue = np.array([14, 90, 253])/255
humon_red = np.array([217, 61, 67])/255


def PlaySkeletonAnimation(
    session=None,
    vidType=1,
    startFrame=0,
    azimuth=-90,
    elevation=-70,
    useOpenPose=True,
    useMediaPipe=False,
    useDLC=False,
    recordVid = True,
    showAnimation =True,
    ):




    ###  
    ###
    ###      ██    ██ ██████  ██████   █████  ████████ ███████         ███████ ██  ██████  ██    ██ ██████  ███████ 
    ###      ██    ██ ██   ██ ██   ██ ██   ██    ██    ██              ██      ██ ██       ██    ██ ██   ██ ██      
    ###      ██    ██ ██████  ██   ██ ███████    ██    █████           █████   ██ ██   ███ ██    ██ ██████  █████   
    ###      ██    ██ ██      ██   ██ ██   ██    ██    ██              ██      ██ ██    ██ ██    ██ ██   ██ ██      
    ###       ██████  ██      ██████  ██   ██    ██    ███████         ██      ██  ██████   ██████  ██   ██ ███████ 
    ###                                                                                                               
    def update_figure(frameNum):
        """ 
        Called by matplotlib animator for each frame.
        """
        
        if useOpenPose:
            op_skel_dottos = matplotlib_artist_objs['op_skel_dottos'] 
            op_skel_trajectories = figure_data['openPose_skel_trajectories|mar|fr_dim']

        if useMediaPipe:
            mp_skel_dottos = matplotlib_artist_objs['mp_skel_dottos'] 
            mp_skel_trajectories = figure_data['mediaPipe_skel_trajectories|mar|fr_dim']

        charuco_dottos = matplotlib_artist_objs['charuco_dottos'] 
        charuco_trajectories = figure_data['charuco_trajectories|mar|fr_dim']

        if useDLC:
            dlc_dottos = matplotlib_artist_objs['dlc_dottos'] 
            dlc_trajectories = figure_data['dlc_trajectories|mar|fr_dim']

        #function to update the lines for each 3d body segment
        def update_3d_skel_segments(key, data_fr_mar_xyz):       
            """ 
            updates the Artist of each body segment with the current frame data.
            """    
            split = key.split('_')

            if split[0] == 'op':
                dict_of_segments_idxs = dict_of_openPoseSegmentIdx_dicts[key]
            if split[0] == 'mp':
                dict_of_segments_idxs = dict_of_mediaPipeSegmentIdx_dicts[key]
            elif split[0] == 'dlc':
                dict_of_segments_idxs = dict_of_dlcSegmentIdx_dicts[key]

            try:
                dict_of_artists = matplotlib_artist_objs[key] 
            except:
                print('KeyError:', key)
                

            for thisItem in dict_of_segments_idxs.items():
                segName = thisItem[0]
                segArtist = dict_of_artists[segName]
                segArtist.set_data((data_fr_mar_xyz[frameNum, dict_of_segments_idxs[segName], 0], data_fr_mar_xyz[frameNum, dict_of_segments_idxs[segName], 1]))
                segArtist.set_3d_properties(data_fr_mar_xyz[frameNum, dict_of_segments_idxs[segName], 2])                

        if useOpenPose:    
            update_3d_skel_segments('op_body', openPose_skel_fr_mar_xyz)
            update_3d_skel_segments('op_rHand', openPose_skel_fr_mar_xyz)
            update_3d_skel_segments('op_lHand', openPose_skel_fr_mar_xyz)
            update_3d_skel_segments('op_face', openPose_skel_fr_mar_xyz)

        if useMediaPipe:
            update_3d_skel_segments('mp_body', mediaPipe_skel_fr_mar_xyz)
            update_3d_skel_segments('mp_rHand', mediaPipe_skel_fr_mar_xyz)
            update_3d_skel_segments('mp_lHand', mediaPipe_skel_fr_mar_xyz)

        if useDLC:
            update_3d_skel_segments('dlc_pinkBall', dlc_fr_mar_xyz)
            update_3d_skel_segments('dlc_redBall', dlc_fr_mar_xyz)
            update_3d_skel_segments('dlc_greenBall', dlc_fr_mar_xyz)
            update_3d_skel_segments('dlc_wobbleBoard', dlc_fr_mar_xyz)
            update_3d_skel_segments('dlc_wobbleWheel', dlc_fr_mar_xyz)

            for bb in range(3):
                thisTailArtist = matplotlib_artist_objs['ball_tails'][bb]
            
                thisTailArtist.set_data((dlc_trajectories[bb][frameNum-ballTailLen:frameNum+1, 0], 
                                        dlc_trajectories[bb][frameNum-ballTailLen:frameNum+1, 1])) 

                thisTailArtist.set_3d_properties(dlc_trajectories[bb][frameNum-ballTailLen:frameNum+1, 2]) 
            #dlc data
            marNum = -1
            for thisDotto, thisTraj in zip(dlc_dottos,dlc_trajectories):
                marNum+=1
                # NOTE: there is no .set_data() for 3 dim data...
                thisDotto.set_data(thisTraj[ frameNum, 0:2])
                thisDotto.set_3d_properties(thisTraj[ frameNum, 2])

        #Plots the dots!
        #openpose data
        if useOpenPose:
            marNum = -1
            for thisSkelDotto, thisTraj in zip(op_skel_dottos,op_skel_trajectories):
                marNum+=1
                # NOTE: there is no .set_data() for 3 dim data...
                thisSkelDotto.set_data(thisTraj[ frameNum, 0:2])
                thisSkelDotto.set_3d_properties(thisTraj[ frameNum, 2])
        
        #mediapipe data
        if useMediaPipe:
            marNum = -1
            for thisSkelDotto, thisTraj in zip(mp_skel_dottos,mp_skel_trajectories):
                marNum+=1
                # NOTE: there is no .set_data() for 3 dim data...
                thisSkelDotto.set_data(thisTraj[ frameNum, 0:2])
                thisSkelDotto.set_3d_properties(thisTraj[ frameNum, 2])



        #charuco data        
        for thisDotto, thisTraj in zip(charuco_dottos,charuco_trajectories):

            if frameNum<charuco_trajectories[0].shape[0]:
                thisDotto.set_data(thisTraj[ frameNum, 0:2])
                thisDotto.set_3d_properties(thisTraj[ frameNum, 2])
            else:
                thisDotto.set_color('none')



        
        # function to update the lines for each body segment
        def update_2d_skel_segments(vidNum,key):       
            """ 
            updates the Artist of each body segment with the current frame data.
            """       
            split = key.split('_')
            if split[0] == 'op':
                dict_of_segments_idxs = dict_of_openPoseSegmentIdx_dicts[key]
                dict_of_artists = thisVidOpenPoseArtist_dict[key] 
            elif split[0] == 'mp':
                dict_of_segments_idxs = dict_of_mediaPipeSegmentIdx_dicts[key]
                dict_of_artists = thisVidMediaPipeArtist_dict[key] 


            for thisItem in dict_of_segments_idxs.items():
                segName = thisItem[0]
                segArtist = dict_of_artists[segName]
    
                if split[0] == 'op':
                    xData = openPose_nCams_nFrames_nImgPts_XYC[vidNum, frameNum, dict_of_segments_idxs[segName],0]
                    yData = openPose_nCams_nFrames_nImgPts_XYC[vidNum, frameNum, dict_of_segments_idxs[segName],1]
                elif split[0] == 'mp':
                    xData = mediaPipe_nCams_nFrames_nImgPts_XYC[vidNum, frameNum, dict_of_segments_idxs[segName],0]
                    yData = mediaPipe_nCams_nFrames_nImgPts_XYC[vidNum, frameNum, dict_of_segments_idxs[segName],1]

                xDataMasked = np.ma.masked_array(xData, mask=(xData==0))
                yDataMasked = np.ma.masked_array(yData, mask=(xData==0))
                segArtist.set_data(xDataMasked,yDataMasked)
                
            
        #update video frames 
        for vidNum, thisVidArtist in enumerate(vidAristList):
            vidCapObjList[vidNum].set(cv2.CAP_PROP_POS_FRAMES, frameNum)
            success, image = vidCapObjList[vidNum].read()

            if success:
                thisVidArtist.set_array(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
            else:
                vidCapObjList[vidNum].set(cv2.CAP_PROP_POS_FRAMES, 0)
                success, image = vidCapObjList[vidNum].read()

            if useOpenPose:
                thisVidOpenPoseArtist_dict = list_of_vidOpenPoseArtistdicts[vidNum]
                update_2d_skel_segments(vidNum,'op_body')
                update_2d_skel_segments(vidNum,'op_rHand')
                update_2d_skel_segments(vidNum,'op_lHand')
                update_2d_skel_segments(vidNum,'op_face')
            
            if useMediaPipe:
                thisVidMediaPipeArtist_dict = list_of_vidMediaPipeArtistdicts[vidNum]
                update_2d_skel_segments(vidNum,'mp_body')
                update_2d_skel_segments(vidNum,'mp_rHand')
                update_2d_skel_segments(vidNum,'mp_lHand')

            if useDLC: #JSM NOTE - this one won't work, we'll need to fix it the next time we get DLC data
                thisVidDLCArtist[0].set_data(  dlcData_nCams_nFrames_nImgPts_XYC[vidNum,frameNum,:,0], 
                                        dlcData_nCams_nFrames_nImgPts_XYC[vidNum,frameNum,:,1],)


        #shift timeseries xlims along with the frameNum
        xTimeSeriesAx.set_xlim([(frameNum/fps)-timeRange, (frameNum/fps)+timeRange])
        yTimeSeriesAx.set_xlim([(frameNum/fps)-timeRange, (frameNum/fps)+timeRange])
        for thisArtistKey in xCurrTimeArtists:

            xCurrTimeArtists[thisArtistKey][0].set_xdata(frameNum/fps) 
            yCurrTimeArtists[thisArtistKey][0].set_xdata(frameNum/fps) 
            if not thisArtistKey == 'blackLine':
                xCurrTimeArtists[thisArtistKey][0].set_ydata(xCurrTimeYdata[thisArtistKey][frameNum] ) 
                yCurrTimeArtists[thisArtistKey][0].set_ydata(yCurrTimeYdata[thisArtistKey][frameNum] ) 

        yCurrTimeArtists['frameNumber'].set_x(frameNum/fps+3/fps) 
        yCurrTimeArtists['frameNumber'].set_text('Frame# '+ str(frameNum))
        # if recordVid:
        #     thisFramePath = str(animationFramePath) + '/'+ str(session.sessionID) + "_frame_" + str(frameNum).zfill(6) + ".png"
        #     plt.savefig(thisFramePath)
        #     # fig.suptitle("nSessionID: {}, Frame: {} of {}".format(session.sessionID, frameNum, numFrames), fontsize=10)
        #     # animSlider.set_val(val=frameNum)

    ####  
    ####   ██████ ██████  ███████  █████  ████████ ███████     ███████ ██  ██████  ██    ██ ██████  ███████ 
    ####  ██      ██   ██ ██      ██   ██    ██    ██          ██      ██ ██       ██    ██ ██   ██ ██      
    ####  ██      ██████  █████   ███████    ██    █████       █████   ██ ██   ███ ██    ██ ██████  █████   
    ####  ██      ██   ██ ██      ██   ██    ██    ██          ██      ██ ██    ██ ██    ██ ██   ██ ██      
    ####   ██████ ██   ██ ███████ ██   ██    ██    ███████     ██      ██  ██████   ██████  ██   ██ ███████ 
    ####                                                                                                    
    

    fig = plt.figure(dpi=200)
    plt.ion()


    ###  
    ###   ██████ ██████  ███████  █████  ████████ ███████     ██████        ██████       █████  ██   ██ ██ ███████ 
    ###  ██      ██   ██ ██      ██   ██    ██    ██               ██       ██   ██     ██   ██  ██ ██  ██ ██      
    ###  ██      ██████  █████   ███████    ██    █████        █████  █████ ██   ██     ███████   ███   ██ ███████ 
    ###  ██      ██   ██ ██      ██   ██    ██    ██               ██       ██   ██     ██   ██  ██ ██  ██      ██ 
    ###   ██████ ██   ██ ███████ ██   ██    ██    ███████     ██████        ██████      ██   ██ ██   ██ ██ ███████ 
    ###                                                                                                            
    ###                                                                                                                                                                              

    # ax3d = p3.Axes3D(fig)
    ax3d = fig.add_subplot(projection='3d')
    ax3d.set_position([-.065, .28, .7, .7]) # [left, bottom, width, height])
    ax3d.set_xticklabels([])
    ax3d.set_yticklabels([])
    ax3d.set_zticklabels([])

    # ax3d.set_xlabel('X')
    # ax3d.set_ylabel('Y')
    # ax3d.set_zlabel('Z')
    # ax3d.set_axis_off()


    ax3d.tick_params(length=0) # WHY DOESNT THIS WORK? I HATE THOSE TICK MARKS >:(
    
    smoothData=True
    smoothWinLength = 5
    smoothOrder = 3

    if useOpenPose:
        try:
            openPose_skel_fr_mar_xyz = np.load(session.dataArrayPath / 'openPoseSkel_3d.npy')
            openPose_nCams_nFrames_nImgPts_XYC = np.load(session.dataArrayPath / 'openPoseData_2d.npy')
        except:
            print('No openPose data found.')

        if smoothData:
            for dim in range(openPose_skel_fr_mar_xyz.shape[2]):
                for mm in range(openPose_skel_fr_mar_xyz.shape[1]):
                    openPose_skel_fr_mar_xyz[:,mm,dim] = savgol_filter(openPose_skel_fr_mar_xyz[:,mm,dim], smoothWinLength, smoothOrder)

    if useMediaPipe:
        try:
            mediaPipe_skel_fr_mar_xyz = np.load(session.dataArrayPath / 'mediaPipeSkel_3d.npy')
            mediaPipe_nCams_nFrames_nImgPts_XYC = np.load(session.dataArrayPath / 'mediaPipeData_2d.npy')            
        except:
            print('No mediaPipe data found.')
                
        if smoothData:
            for dim in range(mediaPipe_skel_fr_mar_xyz.shape[2]):
                for mm in range(mediaPipe_skel_fr_mar_xyz.shape[1]):
                    mediaPipe_skel_fr_mar_xyz[:,mm,dim] = savgol_filter(mediaPipe_skel_fr_mar_xyz[:,mm,dim], smoothWinLength, smoothOrder)

    try:
        charuco_fr_mar_xyz = np.load(session.dataArrayPath / 'charuco_3d_points.npy')
    except:
        console.warning('No charuco data found')

    


    figure_data = dict()
    if useOpenPose:
        figure_data['openPose_skel_trajectories|mar|fr_dim'] = [openPose_skel_fr_mar_xyz[:,markerNum,:] for markerNum in range(openPose_skel_fr_mar_xyz.shape[1])]
        figure_data['openPose_skel_fr_mar_xyz'] = openPose_skel_fr_mar_xyz

        dict_of_openPoseSegmentIdx_dicts, dict_of_op_skel_lineColor = formatOpenPoseStickIndices() #these will help us draw body and hands stick figures

    
    if useMediaPipe:
        mediaPipe_trajectories = [mediaPipe_skel_fr_mar_xyz[:,markerNum,:] for markerNum in range(mediaPipe_skel_fr_mar_xyz.shape[1])]
        figure_data['mediaPipe_skel_trajectories|mar|fr_dim'] = [mediaPipe_skel_fr_mar_xyz[:,markerNum,:] for markerNum in range(mediaPipe_skel_fr_mar_xyz.shape[1])]
        figure_data['mediaPipe_skel_fr_mar_xyz'] = mediaPipe_skel_fr_mar_xyz

        dict_of_mediaPipeSegmentIdx_dicts, dict_of_mp_skel_lineColor = formatMediaPipeStickIndices() #these will help us draw body and hands stick figures

    
    charuco_trajectories = [charuco_fr_mar_xyz[:,markerNum,:] for markerNum in range(charuco_fr_mar_xyz.shape[1])]
    figure_data['charuco_trajectories|mar|fr_dim'] = charuco_trajectories
    figure_data['charuco_fr_mar_xyz'] = charuco_fr_mar_xyz


    if useDLC:
        dlc_fr_mar_xyz = np.load(session.dataArrayPath / "deepLabCut_3d.npy")
        dlcData_nCams_nFrames_nImgPts_XYC = np.load(session.dataArrayPath / "deepLabCutData_2d.npy")
        
        for mm in range(dlc_fr_mar_xyz.shape[1]):
            for dim in range(dlc_fr_mar_xyz.shape[2]):
                dlc_fr_mar_xyz[:,mm,dim] = savgol_filter(dlc_fr_mar_xyz[:,mm,dim], 11, 3)

        ballTailLen = 6
        if ballTailLen > startFrame:
            startFrame = ballTailLen+1 #gotta do this, or the tail will index negative frames (without requiring annoying checks later)
    
        dlc_trajectories = [dlc_fr_mar_xyz[:,markerNum,:] for markerNum in range(dlc_fr_mar_xyz.shape[1])]    
        figure_data['dlc_trajectories|mar|fr_dim'] = dlc_trajectories
        figure_data['dlc_fr_mar_xyz'] = dlc_fr_mar_xyz

    
    
    
    def build_3d_segment_artist_dict(data_fr_mar_xyz,
                                     dict_of_list_of_segment_idxs, 
                                     segColor = 'k', 
                                     lineWidth = 1, 
                                     lineStyle = '-', 
                                     markerType = None, 
                                     marSize = 12, 
                                     markerEdgeColor = 'k',):       
        """ 
        Builds a dictionary of line artists for each 3D body segment.
        """       
        segNames = list(dict_of_list_of_segment_idxs)



        dict_of_artist_objects = dict()
        for segNum, segName in enumerate(segNames):

            #determine color of segment, based on class of 'segColor' input
            if isinstance(segColor, str):
                thisRGBA = segColor
            elif isinstance(segColor, np.ndarray):
                thisRGBA = segColor
            elif isinstance(segColor, dict):
                thisRGBA = segColor[segName]
            elif isinstance(segColor, list):
                try:
                    thisRGBA = segColor[segNum]
                except:
                    print('Not enough colors provided, using Black instead')
                    thisRGBA = 'k'
            else:
                thisRGBA = 'k'

            if isinstance(segName, str):
                idxsOG = dict_of_list_of_segment_idxs[segName]
            else:
                idxsOG

            if isinstance(idxsOG, int) or isinstance(idxsOG, float): 
                idxs = [idxsOG]
            elif isinstance(idxsOG, dict):
                idxs = idxsOG[0]
            else:
                idxs = idxsOG.copy()


            dict_of_artist_objects[segName]  = ax3d.plot(
                                                    data_fr_mar_xyz[startFrame, idxs ,0], 
                                                    data_fr_mar_xyz[startFrame, idxs ,1], 
                                                    data_fr_mar_xyz[startFrame, idxs ,2],
                                                    linestyle=lineStyle,
                                                    linewidth=lineWidth,
                                                    markerSize = marSize,
                                                    marker = markerType,
                                                    color = thisRGBA,
                                                    markeredgecolor = markerEdgeColor,
                                                    )[0]
        return dict_of_artist_objects

    

    matplotlib_artist_objs = dict()

    if useOpenPose:
        matplotlib_artist_objs['op_body'] = build_3d_segment_artist_dict(openPose_skel_fr_mar_xyz, dict_of_openPoseSegmentIdx_dicts['op_body'], segColor = dict_of_op_skel_lineColor)
        matplotlib_artist_objs['op_rHand'] = build_3d_segment_artist_dict(openPose_skel_fr_mar_xyz, dict_of_openPoseSegmentIdx_dicts['op_rHand'], segColor=np.append(humon_red, 1), markerType='.', markerEdgeColor = humon_red, lineWidth=1, marSize = 2)
        matplotlib_artist_objs['op_lHand'] = build_3d_segment_artist_dict(openPose_skel_fr_mar_xyz, dict_of_openPoseSegmentIdx_dicts['op_lHand'], segColor=np.append(humon_blue, 1), markerType='.', markerEdgeColor = humon_blue, lineWidth=1, marSize = 2)
        matplotlib_artist_objs['op_face'] = build_3d_segment_artist_dict(openPose_skel_fr_mar_xyz, dict_of_openPoseSegmentIdx_dicts['op_face'], segColor='k', lineWidth=.5)

    if useMediaPipe:
        matplotlib_artist_objs['mp_body'] = build_3d_segment_artist_dict(mediaPipe_skel_fr_mar_xyz, dict_of_mediaPipeSegmentIdx_dicts['mp_body'], segColor = dict_of_mp_skel_lineColor)
        matplotlib_artist_objs['mp_rHand'] = build_3d_segment_artist_dict(mediaPipe_skel_fr_mar_xyz, dict_of_mediaPipeSegmentIdx_dicts['mp_rHand'], segColor=np.append(humon_red, 1), markerType='.', markerEdgeColor = humon_red, lineWidth=1, marSize = 2)
        matplotlib_artist_objs['mp_lHand'] = build_3d_segment_artist_dict(mediaPipe_skel_fr_mar_xyz, dict_of_mediaPipeSegmentIdx_dicts['mp_lHand'], segColor=np.append(humon_blue, 1), markerType='.', markerEdgeColor = humon_blue, lineWidth=1, marSize = 2)

    if useDLC:
        dict_of_dlcSegmentIdx_dicts = dict()
        dict_of_dlcSegmentIdx_dicts['pinkBall'] = {'pinkBall': [0]}
        dict_of_dlcSegmentIdx_dicts['greenBall'] = {'greenBall': [1]}
        dict_of_dlcSegmentIdx_dicts['redBall'] = {'redBall': [2]}
        dict_of_dlcSegmentIdx_dicts['wobbleBoard'] = {'wobbleboard':[3,4,5,6,3,7,5,4,7,6]}
        dict_of_dlcSegmentIdx_dicts['wobbleWheel'] = {'wobbleWheel':[8,9]}
    
  
        ballColor = ['darkviolet', 'forestgreen', 'xkcd:goldenrod']
        matplotlib_artist_objs['pinkBall'] = build_3d_segment_artist_dict(dlc_fr_mar_xyz, dict_of_dlcSegmentIdx_dicts['pinkBall'], segColor=ballColor[0], markerType = 'o', marSize = 6, markerEdgeColor = 'indigo')
        matplotlib_artist_objs['redBall'] = build_3d_segment_artist_dict(dlc_fr_mar_xyz, dict_of_dlcSegmentIdx_dicts['redBall'], segColor=ballColor[2], markerType = 'o', marSize = 6, markerEdgeColor ='orangered')
        matplotlib_artist_objs['greenBall'] = build_3d_segment_artist_dict(dlc_fr_mar_xyz, dict_of_dlcSegmentIdx_dicts['greenBall'], segColor=ballColor[1], markerType = 'o', marSize = 6, markerEdgeColor ='darkgreen')

        matplotlib_artist_objs['wobbleBoard'] = build_3d_segment_artist_dict(dlc_fr_mar_xyz, dict_of_dlcSegmentIdx_dicts['wobbleBoard'], segColor='k', lineWidth=1)
        matplotlib_artist_objs['wobbleWheel'] = build_3d_segment_artist_dict(dlc_fr_mar_xyz, dict_of_dlcSegmentIdx_dicts['wobbleWheel'], segColor='k', lineWidth=1)

    if useOpenPose:
        op_skel_dottos = []
        for mm in range(67): #getcher dottos off my face!
            thisTraj = openPose_skel_fr_mar_xyz[:, mm, :]
            if mm==15:
                col = 'r'
                markerSize = 2
            elif mm == 16:
                col = 'b'
                markerSize = 2
            else:
                col = 'k'
                markerSize = 1

            thisDotto =ax3d.plot(thisTraj[0, 0:1], thisTraj[1, 0:1], thisTraj[2, 0:1][0],'.', markersize=markerSize, color = col)
            op_skel_dottos.append(thisDotto[0])

            matplotlib_artist_objs['op_skel_dottos'] = op_skel_dottos

    if useMediaPipe:
        matplotlib_artist_objs['mp_skel_dottos'] = [ax3d.plot(thisTraj[0, 0:1], thisTraj[1, 0:1], thisTraj[2, 0:1],'m.', markersize=1)[0] for thisTraj in mediaPipe_trajectories]

    matplotlib_artist_objs['charuco_dottos'] = [ax3d.plot(thisTraj[0, 0:1], thisTraj[1, 0:1], thisTraj[2, 0:1],'m.', markersize=1)[0] for thisTraj in charuco_trajectories]

    
    if useDLC:
        matplotlib_artist_objs['dlc_dottos'] = [ax3d.plot(thisTraj[0, 0:1], thisTraj[1, 0:1], thisTraj[2, 0:1],'b.', markersize=1)[0] for thisTraj in dlc_trajectories]

        matplotlib_artist_objs['ball_tails'] = [ax3d.plot( 
                                                dlc_trajectories[bb][startFrame-ballTailLen:startFrame, 0],
                                                dlc_trajectories[bb][startFrame-ballTailLen:startFrame, 1],
                                                dlc_trajectories[bb][startFrame-ballTailLen:startFrame, 2],
                                                '-', color=ballColor[bb])[0] for bb in range(3)]
    

    if useOpenPose:
        numFrames = openPose_skel_fr_mar_xyz.shape[0]
        mx = np.nanmean(openPose_skel_fr_mar_xyz[int(numFrames/2),:,0])
        my = np.nanmean(openPose_skel_fr_mar_xyz[int(numFrames/2),:,1])
        mz = np.nanmean(openPose_skel_fr_mar_xyz[int(numFrames/2),:,2])

    if useMediaPipe:
        numFrames = mediaPipe_skel_fr_mar_xyz.shape[0]
        mx = np.nanmean(mediaPipe_skel_fr_mar_xyz[int(numFrames/2),:,0])
        my = np.nanmean(mediaPipe_skel_fr_mar_xyz[int(numFrames/2),:,1])
        mz = np.nanmean(mediaPipe_skel_fr_mar_xyz[int(numFrames/2),:,2])

    
    if np.isnan(mx) or np.isnan(my) or np.isnan(mz):
        mx = 0
        my = 0
        mz = 0
    # groundX = np.arange(mx-100,mx+100)
    # groundZ = np.arange(my-100,my+100)
    # groundXX, groundZZ = np.meshgrid(groundX, groundZ)
    # groundYY = np.zeros_like(groundXX)
    # groundMesh = ax3d.plot_surface(groundXX, groundZZ, groundYY, color='k', alpha=.5)

    axRange = 1000#session.board.square_length * 10

    # Setting the axes properties
    ax3d.set_xlim3d([mx-axRange, mx+axRange])
    ax3d.set_ylim3d([my-axRange, my+axRange])
    ax3d.set_zlim3d([mz-axRange, mz+axRange])
    
    tit1 = fig.text(.5, .96, "The FreeMoCap Project", fontsize=15, horizontalalignment='center')
    tit2 = fig.text(.5, .92, "Session: {}".format(session.sessionID), fontsize=8, horizontalalignment='center')



    ax3d.view_init(azim=azimuth, elev=elevation)
    


    ###   
    ###   
    ###   ███    ███  █████  ██   ██ ███████     ██    ██ ██ ██████  ███████  ██████       █████  ██   ██ ███████ ███████ 
    ###   ████  ████ ██   ██ ██  ██  ██          ██    ██ ██ ██   ██ ██      ██    ██     ██   ██  ██ ██  ██      ██      
    ###   ██ ████ ██ ███████ █████   █████       ██    ██ ██ ██   ██ █████   ██    ██     ███████   ███   █████   ███████ 
    ###   ██  ██  ██ ██   ██ ██  ██  ██           ██  ██  ██ ██   ██ ██      ██    ██     ██   ██  ██ ██  ██           ██ 
    ###   ██      ██ ██   ██ ██   ██ ███████       ████   ██ ██████  ███████  ██████      ██   ██ ██   ██ ███████ ███████ 
    ###                                                                                                                   
    ###                                                                                                                   
    

    syncedVidPathListAll = list(sorted(session.syncedVidPath.glob('*.mp4')))
    numVids = len(syncedVidPathListAll)

    syncedVidPathList  = syncedVidPathListAll.copy()

    vidAxesList = []
    vidAristList = []
    vidCapObjList = []

    list_of_vidOpenPoseArtistdicts = []
    list_of_vidMediaPipeArtistdicts = []
    
    if useDLC: vidDLCArtistList = []    

    def build_2d_segment_artist_dict(vidNum, data_nCams_nFrames_nImgPts_XYC, dict_of_list_of_segment_idxs, segColor = 'k', lineWidth = 1, lineStyle = '-'):       
        """ 
        Builds a dictionary of line artists for each body 2d segment.
        """       
        segNames = list(dict_of_list_of_segment_idxs)

        dict_of_artist_objects = dict()
        for segNum, segName in enumerate(segNames):

            #determine color of segment, based on class of 'segColor' input
            if isinstance(segColor, str):
                thisRGBA = segColor
            elif isinstance(segColor, np.ndarray):
                thisRGBA = segColor
            elif isinstance(segColor, dict):
                thisRGBA = segColor[segName].copy()
                thisRGBA[-1] = .75
            elif isinstance(segColor, list):
                try:
                    thisRGBA = segColor[segNum]
                except:
                    print('Not enough colors provided, using Black instead')
                    thisRGBA = 'k'
            else:
                thisRGBA = 'k'

            xData = data_nCams_nFrames_nImgPts_XYC[vidNum, startFrame, dict_of_list_of_segment_idxs[segName],0]
            yData = data_nCams_nFrames_nImgPts_XYC[vidNum, startFrame, dict_of_list_of_segment_idxs[segName],1]

            # #make NaN's invisible (i thought they already would be but???!!!!)
            # thisRGBAall = np.tile(thisRGBA,(len(xData),1))
            # thisRGBAall[np.isnan(xData),3] = 0
            
            xDataMasked  = np.ma.masked_where(xData, np.isnan(xData))
            yDataMasked  = np.ma.masked_where(yData, np.isnan(yData))

            dict_of_artist_objects[segName]  = thisVidAxis.plot(
                                                    xDataMasked,
                                                    yDataMasked,
                                                    linestyle=lineStyle,
                                                    linewidth=lineWidth,
                                                    color = thisRGBA,
                                                    )[0]
        return dict_of_artist_objects

    # vidImageGrid = ImageGrid(fig, [.45, 0.05, .54, .9], nrows_ncols = (2,2), axes_pad = 0.1) # https://matplotlib.org/stable/gallery/axes_grid1/simple_axesgrid.html
    if numVids < 4:
        numRows = numVids
        numCols = 1
    else:
        numRows = int(np.ceil(numVids/2))
        numCols = 2


    vidAxGridSpec = fig.add_gridspec(numRows, numCols, left=.55, bottom=.025, right=.9, top=.9, wspace=.1, hspace=.1)


    for thisVidNum, thisVidPath in enumerate(syncedVidPathList):
        #make subplot for figure (and set position)

        if numVids < 4:
            thisVidAxis = fig.add_subplot(vidAxGridSpec[thisVidNum])
        elif numVids %2 > 0: #if odd number videos plot the first vid across 2 spots
            if thisVidNum==0:
                thisVidAxis = fig.add_subplot(vidAxGridSpec[0,:])
            else:
                thisVidAxis = fig.add_subplot(vidAxGridSpec[thisVidNum+1])
        else: 
            thisVidAxis = fig.add_subplot(vidAxGridSpec[thisVidNum])


        thisVidAxis.set_axis_off()

        vidAxesList.append(thisVidAxis)

        #create video capture object
        thisVidCap = cv2.VideoCapture(str(thisVidPath))
        

        #create artist object for each video 
        success, image  = thisVidCap.read()

        assert success==True, "{} - failed to load an image".format(thisVidPath.stem) #make sure we have a frame

        vidAristList.append(thisVidAxis.imshow(cv2.cvtColor(image, cv2.COLOR_BGR2RGB)))
        vidCapObjList.append(thisVidCap)



        if useOpenPose:
            vidOpenPoseArtist_dict = dict()            
            vidOpenPoseArtist_dict['op_body'] = build_2d_segment_artist_dict(thisVidNum, openPose_nCams_nFrames_nImgPts_XYC, dict_of_openPoseSegmentIdx_dicts['op_body'], segColor = dict_of_op_skel_lineColor)
            vidOpenPoseArtist_dict['op_rHand'] = build_2d_segment_artist_dict(thisVidNum, openPose_nCams_nFrames_nImgPts_XYC, dict_of_openPoseSegmentIdx_dicts['op_rHand'], segColor=np.append(humon_red, .75), lineWidth=.5)
            vidOpenPoseArtist_dict['op_lHand'] = build_2d_segment_artist_dict(thisVidNum, openPose_nCams_nFrames_nImgPts_XYC, dict_of_openPoseSegmentIdx_dicts['op_lHand'], segColor=np.append(humon_blue, .75), lineWidth=.5)
            vidOpenPoseArtist_dict['op_face'] = build_2d_segment_artist_dict(thisVidNum, openPose_nCams_nFrames_nImgPts_XYC, dict_of_openPoseSegmentIdx_dicts['op_face'], segColor = np.array([1.,1.,1.,1]), lineWidth=.25)
            list_of_vidOpenPoseArtistdicts.append(vidOpenPoseArtist_dict)

        if useMediaPipe:
            vidMediaPipeArtist_dict = dict()            
            vidMediaPipeArtist_dict['mp_body'] = build_2d_segment_artist_dict(thisVidNum, mediaPipe_nCams_nFrames_nImgPts_XYC, dict_of_mediaPipeSegmentIdx_dicts['mp_body'], segColor = 'g')
            vidMediaPipeArtist_dict['mp_rHand'] = build_2d_segment_artist_dict(thisVidNum, mediaPipe_nCams_nFrames_nImgPts_XYC, dict_of_mediaPipeSegmentIdx_dicts['mp_rHand'], segColor=np.append(humon_red, .75), lineWidth=.5)
            vidMediaPipeArtist_dict['mp_lHand'] = build_2d_segment_artist_dict(thisVidNum, mediaPipe_nCams_nFrames_nImgPts_XYC, dict_of_mediaPipeSegmentIdx_dicts['mp_lHand'], segColor=np.append(humon_blue, .75), lineWidth=.5)
            list_of_vidMediaPipeArtistdicts.append(vidMediaPipeArtist_dict)


        if useDLC:
            
            vidDLCArtistList.append(thisVidAxis.plot(
                                    dlcData_nCams_nFrames_nImgPts_XYC[thisVidNum,startFrame,:,0], 
                                    dlcData_nCams_nFrames_nImgPts_XYC[thisVidNum,startFrame,:,1],
                                    linestyle='none',
                                    marker = '.',
                                    markersize = 1,
                                    color='w',
                                    markerfacecolor='none' ))

    ###               
    ###   
    ###   ███    ███  █████  ██   ██ ███████     ██      ██ ███    ██ ███████     ██████  ██       ██████  ████████      █████  ██   ██ ███████ ███████ 
    ###   ████  ████ ██   ██ ██  ██  ██          ██      ██ ████   ██ ██          ██   ██ ██      ██    ██    ██        ██   ██  ██ ██  ██      ██      
    ###   ██ ████ ██ ███████ █████   █████       ██      ██ ██ ██  ██ █████       ██████  ██      ██    ██    ██        ███████   ███   █████   ███████ 
    ###   ██  ██  ██ ██   ██ ██  ██  ██          ██      ██ ██  ██ ██ ██          ██      ██      ██    ██    ██        ██   ██  ██ ██  ██           ██ 
    ###   ██      ██ ██   ██ ██   ██ ███████     ███████ ██ ██   ████ ███████     ██      ███████  ██████     ██        ██   ██ ██   ██ ███████ ███████ 
    ###                                                                                                                                                 
    ###                                                                                                                                                 


    fps=30 #NOTE - This should be saved in the session Class some how
    time = np.arange(0, numFrames)/fps
    if useMediaPipe:
        rHandIdx = 20
        lHandIdx = 19
        rHandX = mediaPipe_skel_fr_mar_xyz[:,rHandIdx,0]/1000 #convert mm to meters
        rHandY = -mediaPipe_skel_fr_mar_xyz[:,rHandIdx,1]/1000
        lHandX = mediaPipe_skel_fr_mar_xyz[:,lHandIdx,0]/1000
        lHandY = -mediaPipe_skel_fr_mar_xyz[:,lHandIdx,1]/1000

    if useOpenPose:
        rHandIdx = 4
        lHandIdx = 7
        rHandX = openPose_skel_fr_mar_xyz[:,rHandIdx,0]/1000 #convert mm to meters
        rHandY = -openPose_skel_fr_mar_xyz[:,rHandIdx,1]/1000
        lHandX = openPose_skel_fr_mar_xyz[:,lHandIdx,0]/1000
        lHandY = -openPose_skel_fr_mar_xyz[:,lHandIdx,1]/1000

    
    linePlotWidth = .45
    linePlotHeight = .14

    yTimeSeriesAx = fig.add_subplot(position=[.07, .225, linePlotWidth, linePlotHeight])
    yTimeSeriesAx.set_title('Hand Position vs Time', fontsize = 7, pad=2)

    yTimeSeriesAx.plot(time, rHandY, color=humon_red, linewidth=.75, label='Right Hand')
    yTimeSeriesAx.plot(time, lHandY, color=humon_blue, linewidth=.75, label='Left Hand')
    
    timeRange = 3
    ylimRange = .6

    yCurrTimeArtists = dict()
    yCurrTimeYdata = dict()
    yCurrTimeArtists['blackLine']  = yTimeSeriesAx.plot([startFrame/fps, startFrame/fps], [-ylimRange, ylimRange*2], color='k', linewidth=1)

    yCurrTimeArtists['rHandDot']  = yTimeSeriesAx.plot([startFrame/fps], [rHandY[startFrame]], markeredgecolor=humon_red, markerfacecolor = 'k',marker='o', markersize=3)
    yCurrTimeYdata['rHandDot'] = rHandY
    yCurrTimeArtists['lHandDot']  = yTimeSeriesAx.plot([startFrame/fps], [lHandY[startFrame]], markeredgecolor=humon_blue, markerfacecolor = 'k', marker='o', markersize=3)
    yCurrTimeYdata['lHandDot'] = lHandY
    
    
    yCurrTimeArtists['frameNumber'] = yTimeSeriesAx.text(startFrame/fps+3/fps, ylimRange*.7, "Frame# " + str(startFrame), fontsize=6)

    yTimeSeriesAx.tick_params(labelsize=6, direction='in', width=.5)
    yTimeSeriesAx.tick_params( pad=2)
    

    yTimeSeriesAx.set_ylabel('Vertical (m)', fontsize=7, labelpad=3)
    yTimeSeriesAx.set_ylim([-ylimRange, ylimRange])
    yTimeSeriesAx.set_xlim([(startFrame/fps)-timeRange, (startFrame/fps)+timeRange])

    

    for axis in ['top','bottom','left','right']:
        yTimeSeriesAx.spines[axis].set_linewidth(0.5)



    yTimeSeriesAx.legend(loc='upper left', fontsize=6)



    xTimeSeriesAx = fig.add_subplot(position=[.07, 0.05, linePlotWidth, linePlotHeight])
    
    xTimeSeriesAx.plot(time, rHandX, color=humon_red, linewidth=.75, label='Right Hand')
    xTimeSeriesAx.plot(time, lHandX, color=humon_blue, linewidth=.75, label='Left Hand')
    


    xCurrTimeArtists = dict()
    xCurrTimeYdata = dict()
    xCurrTimeArtists['blackLine']  = xTimeSeriesAx.plot([startFrame/fps, startFrame/fps], [-ylimRange, ylimRange], color='k', linewidth=1)
    xCurrTimeArtists['rHandDot']  = xTimeSeriesAx.plot([startFrame/fps], [rHandX[startFrame]], markeredgecolor=humon_red, markerfacecolor = 'k',marker='o', markersize=3)
    xCurrTimeYdata['rHandDot'] = rHandX
    xCurrTimeArtists['lHandDot']  = xTimeSeriesAx.plot([startFrame/fps], [lHandX[startFrame]], markeredgecolor=humon_blue, markerfacecolor = 'k', marker='o', markersize=3)
    xCurrTimeYdata['lHandDot'] = lHandX



    xTimeSeriesAx.tick_params(labelsize=6, direction='in', width=.5)
    xTimeSeriesAx.tick_params( pad=2)
    
    xTimeSeriesAx.set_ylabel('Left/Right (m)', fontsize=7, labelpad=0)
    xTimeSeriesAx.set_xlabel('Time(sec)', fontsize=8, labelpad=0)

    xTimeSeriesAx.set_ylim([-ylimRange, ylimRange])

    xTimeSeriesAx.set_xlim([(startFrame/fps)-timeRange, (startFrame/fps)+timeRange])

    
    for axis in ['top','bottom','left','right']:
        xTimeSeriesAx.spines[axis].set_linewidth(0.5)



    # #slider
    # axControls = fig.add_subplot(2,1,2)
    # axControls.set_position([0.25, 0.01, 0.6, 0.05])
    # animSlider = Slider(
    #     ax=axControls,
    #     label="FrameNum",
    #     valmin=10,
    #     valmax=numFrames,
    #     valinit=10,
    #     orientation="horizontal"
    # )

    # thisVidAxis.text(-565,1440, 'Music - artist: Neon Exdeath, song: Meowmaline, album: Jewel Tones',color=humon_green, fontsize=6)
    thisVidAxis.text(0,thisVidCap.get(cv2.CAP_PROP_FRAME_HEIGHT)*1.06, 'github.com/jonmatthis/freemocap || freemocap.org',color=humon_green, fontsize=6)

    logoAx = fig.add_subplot(position=[.85, .85, .15, .15])
    logoIm = cv2.imread(r'logo\fmc-logo-black-border-white-bkgd.png') #JSM NOTE - THis doesn't work, but SOMETIMES IT DOES?!
    if logoIm is not None:  logoAx.imshow(cv2.cvtColor(logoIm, cv2.COLOR_BGR2RGB))
    logoAx.axis('off')
    
    # if recordVid:
    #     animationFramePath = session.sessionPath / "animationFrames"
    #     animationFramePath.mkdir(parents=True, exist_ok=True)

    # Creating the Animation object
    line_animation = animation.FuncAnimation(fig, update_figure, range(startFrame,numFrames), fargs=(),
                                    interval=1, blit=False)


    
    if recordVid:
        # vidSavePath = '{}_animVid.mp4'.format(str(session.sessionPath / session.sessionID))
        gifSavePath = '{}_animVid.gif'.format(str(session.sessionPath / session.sessionID)) #NOTE - saving as a gif first then converting to MP4, because saving directly to MP4 requires users to install ffmpeg (independently of the pip install)
        with console.status('Saving animation for {}'.format(session.sessionID)):
            # Writer = animation.writers['ffmpeg']
            Writer = animation.writers['pillow']            
            writer = Writer(fps=fps, metadata=dict(artist='FreeMoCap'))#, bitrate=1800)
            line_animation.save(gifSavePath, writer = writer)
            # Writer = animation.FFMpegWriter(fps=30, metadata=dict(artist='FreeMoCap', comment=session.sessionID), bitrate=1800)
            # vidSavePath = '{}_outVid.mp4'.format(str(session.sessionPath / session.sessionID))
            # Writer.saving(fig = fig, outfile=vidSavePath, dpi=150)
        
        gif_filepath = mp.VideoFileClip(gifSavePath)
        gif_filepath.write_videofile(gifSavePath.replace('.gif','.mp4'))
        os.remove(gifSavePath)


    try:
        if showAnimation:
            with console.status('Playing Skeleton animation! Close the `matplotlib` window to continue...'):
                plt.pause(0.1)
                plt.draw()
    except:
        pass



  


    ### 
    ###    
    ###    ███████  ██████  ██████  ███    ███  █████  ████████      ██████  ██████  ███████ ███    ██ ██████   ██████  ███████ ███████     ███████ ██   ██ ███████ ██      
    ###    ██      ██    ██ ██   ██ ████  ████ ██   ██    ██        ██    ██ ██   ██ ██      ████   ██ ██   ██ ██    ██ ██      ██          ██      ██  ██  ██      ██      
    ###    █████   ██    ██ ██████  ██ ████ ██ ███████    ██        ██    ██ ██████  █████   ██ ██  ██ ██████  ██    ██ ███████ █████       ███████ █████   █████   ██      
    ###    ██      ██    ██ ██   ██ ██  ██  ██ ██   ██    ██        ██    ██ ██      ██      ██  ██ ██ ██      ██    ██      ██ ██               ██ ██  ██  ██      ██      
    ###    ██       ██████  ██   ██ ██      ██ ██   ██    ██         ██████  ██      ███████ ██   ████ ██       ██████  ███████ ███████     ███████ ██   ██ ███████ ███████ 
    ###                                                                                                                                                                         
    ###                                                                                                                                                                         
def formatOpenPoseStickIndices():
    """
    generate dictionary of arrays, each containing the 'connect-the-dots' order to draw a given body segment
    
    returns:
    openPoseBodySegmentIds= a dictionary of arrays containing indices of individual body segments (Note, a lot of markerless mocap comp sci types like to say 'pose' instead of 'body'. They also use 'pose' to refer to camera 6 DoF position sometimes. Comp sci is frustrating like that lol)
    openPoseHandIds = a dictionary of arrays containing indices of individual hand segments, along with offset to know where to start in the 'skel_fr_mar_xyz.shape[1]' part of the array
    dict_of_op_skel_lineColor = a dictionary of arrays, each containing the color (RGBA) to use for a given body segment
    """
    dict_of_openPoseSegmentIdx_dicts = dict()

    #make body dictionary
    openPoseBodySegmentIds = dict()
    openPoseBodySegmentIds['op_head'] = [17, 15, 0, 1,0, 16, 18, ]
    openPoseBodySegmentIds['op_spine'] = [1,8,5,1, 2, 12, 8, 9, 5, 1, 2, 8]
    openPoseBodySegmentIds['op_rArm'] = [1, 2, 3, 4, ]
    openPoseBodySegmentIds['op_lArm'] = [1, 5, 6, 7, ]
    openPoseBodySegmentIds['op_rLeg'] = [8, 9, 10, 11, 22, 23, 11, 24, ]
    openPoseBodySegmentIds['op_lLeg'] = [8,12, 13, 14, 19, 20, 14, 21,]
    dict_of_openPoseSegmentIdx_dicts['op_body'] = openPoseBodySegmentIds


    #make colors dictionary
    openPoseBodyColor = np.array([50, 89, 0])/255
    openPoseRightColor = np.array([230, 50, 0])/255
    openPoseLeftColor = np.array([0, 50, 230])/255
    
    dict_of_op_skel_lineColor = dict()
    dict_of_op_skel_lineColor['op_head'] = np.append(openPoseBodyColor, 0.5)
    dict_of_op_skel_lineColor['op_spine'] = np.append(openPoseBodyColor, 1)
    dict_of_op_skel_lineColor['op_rArm'] = np.append(openPoseRightColor, 1)
    dict_of_op_skel_lineColor['op_lArm'] = np.append(openPoseLeftColor, 1)
    dict_of_op_skel_lineColor['op_rLeg'] = np.append(openPoseRightColor, 1)
    dict_of_op_skel_lineColor['op_lLeg'] = np.append(openPoseLeftColor, 1)


    # Make some handy maps ;D
    openPoseHandIds = dict()
    rHandIDstart = 25
    lHandIDstart = rHandIDstart + 21

    openPoseHandIds['op_thumb'] = np.array([0, 1, 2, 3, 4,  ]) 
    openPoseHandIds['op_index'] = np.array([0, 5, 6, 7, 8, ])
    openPoseHandIds['op_bird']= np.array([0, 9, 10, 11, 12, ])
    openPoseHandIds['op_ring']= np.array([0, 13, 14, 15, 16, ])
    openPoseHandIds['op_pinky'] = np.array([0, 17, 18, 19, 20, ])
    

    rHand_dict = copy.deepcopy(openPoseHandIds.copy()) #copy.deepcopy() is necessary to make sure the dicts are independent of each other
    lHand_dict = copy.deepcopy(rHand_dict)

    for key in rHand_dict: 
        rHand_dict[key] += rHandIDstart 
        lHand_dict[key] += lHandIDstart 

    dict_of_openPoseSegmentIdx_dicts['op_rHand'] = rHand_dict
    dict_of_openPoseSegmentIdx_dicts['op_lHand'] = lHand_dict

    
    #how to face --> :D <--
    openPoseFaceIDs = dict()
    faceIDStart = 67
    #define face parts
    openPoseFaceIDs['op_jaw'] = np.arange(0,16) + faceIDStart 
    openPoseFaceIDs['op_rBrow'] = np.arange(17,21) + faceIDStart
    openPoseFaceIDs['op_lBrow'] = np.arange(22,26) + faceIDStart
    openPoseFaceIDs['op_noseRidge'] = np.arange(27,30) + faceIDStart
    openPoseFaceIDs['op_noseBot'] = np.arange(31,35) + faceIDStart
    openPoseFaceIDs['op_rEye'] = np.concatenate((np.arange(36,41), [36])) + faceIDStart
    openPoseFaceIDs['op_lEye'] = np.concatenate((np.arange(42,47), [42])) + faceIDStart    
    openPoseFaceIDs['op_upperLip'] = np.concatenate((np.arange(48,54), np.flip(np.arange(60, 64)), [48])) + faceIDStart
    openPoseFaceIDs['op_lowerLip'] = np.concatenate(([60], np.arange(64,67), np.arange(54, 59), [48], [60])) + faceIDStart
    openPoseFaceIDs['op_rPupil'] = np.array([68]) + faceIDStart
    openPoseFaceIDs['op_lPupil'] = np.array([69]) + faceIDStart #nice

    dict_of_openPoseSegmentIdx_dicts['op_face'] = openPoseFaceIDs
    
    return dict_of_openPoseSegmentIdx_dicts, dict_of_op_skel_lineColor



###  
###  ███████  ██████  ██████  ███    ███  █████  ████████     ███    ███ ███████ ██████  ██  █████  ██████  ██ ██████  ███████     ███████ ██   ██ ███████ ██      
###  ██      ██    ██ ██   ██ ████  ████ ██   ██    ██        ████  ████ ██      ██   ██ ██ ██   ██ ██   ██ ██ ██   ██ ██          ██      ██  ██  ██      ██      
###  █████   ██    ██ ██████  ██ ████ ██ ███████    ██        ██ ████ ██ █████   ██   ██ ██ ███████ ██████  ██ ██████  █████       ███████ █████   █████   ██      
###  ██      ██    ██ ██   ██ ██  ██  ██ ██   ██    ██        ██  ██  ██ ██      ██   ██ ██ ██   ██ ██      ██ ██      ██               ██ ██  ██  ██      ██      
###  ██       ██████  ██   ██ ██      ██ ██   ██    ██        ██      ██ ███████ ██████  ██ ██   ██ ██      ██ ██      ███████     ███████ ██   ██ ███████ ███████ 
###                                                                                                                                                                
###                                                                                                                                                                                                                                                                                                                                       
def formatMediaPipeStickIndices():
    """
    generate dictionary of arrays, each containing the 'connect-the-dots' order to draw a given body segment
    
    returns:
    mediaPipeBodySegmentIds= a dictionary of arrays containing indices of individual body segments (Note, a lot of markerless mocap comp sci types like to say 'pose' instead of 'body'. They also use 'pose' to refer to camera 6 DoF position sometimes. Comp sci is frustrating like that lol)
    mediaPipeHandIds = a dictionary of arrays containing indices of individual hand segments, along with offset to know where to start in the 'skel_fr_mar_xyz.shape[1]' part of the array
    dict_of_mp_skel_lineColor = a dictionary of arrays, each containing the color (RGBA) to use for a given body segment
    """
    dict_of_mediaPipeSegmentIdx_dicts = dict()

    #make body dictionary
    mediaPipeBodySegmentIds = dict()
    mediaPipeBodySegmentIds['mp_head'] = [8, 6, 5, 4, 0, 10, 9, 0, 1, 2, 3, 7 ]
    mediaPipeBodySegmentIds['mp_tors'] = [12, 11, 24, 23, 12]
    mediaPipeBodySegmentIds['mp_rArm'] = [12, 14, 16, 18, 20, 16, 22 ]
    mediaPipeBodySegmentIds['mp_lArm'] = [11, 13, 15, 17, 19, 15, 21]
    mediaPipeBodySegmentIds['mp_rLeg'] = [24, 26, 28, 30, 32, 28, ]
    mediaPipeBodySegmentIds['mp_lLeg'] = [23, 25, 27, 29, 31, 27, ]
    dict_of_mediaPipeSegmentIdx_dicts['mp_body'] = mediaPipeBodySegmentIds

    #make colors dictionary
    mediaPipeBodyColor = np.array([125,106,0])/255
    mediaPipeRightColor = np.array([230, 0, 169])/255
    mediaPipeLeftColor = np.array([0, 176, 176])/255
    dict_of_mp_skel_lineColor = dict()
    
    dict_of_mp_skel_lineColor['mp_head'] = np.append(mediaPipeBodyColor, .5)
    dict_of_mp_skel_lineColor['mp_tors'] = np.append(mediaPipeBodyColor, 1)
    dict_of_mp_skel_lineColor['mp_rArm'] = np.append(mediaPipeRightColor, 1)
    dict_of_mp_skel_lineColor['mp_lArm'] = np.append(mediaPipeLeftColor, 1)
    dict_of_mp_skel_lineColor['mp_rLeg'] = np.append(mediaPipeRightColor, 1)
    dict_of_mp_skel_lineColor['mp_lLeg'] = np.append(mediaPipeLeftColor, 1)


    # Make some handy maps ;D
    mediaPipeHandIds = dict()
    rHandIDstart = 33
    lHandIDstart = rHandIDstart + 21

    mediaPipeHandIds['mp_thumb'] = np.array([0, 1, 2, 3, 4,  ]) 
    mediaPipeHandIds['mp_index'] = np.array([0, 5, 6, 7, 8, ])
    mediaPipeHandIds['mp_bird']= np.array([0, 9, 10, 11, 12, ])
    mediaPipeHandIds['mp_ring']= np.array([0, 13, 14, 15, 16, ])
    mediaPipeHandIds['mp_pinky'] = np.array([0, 17, 18, 19, 20, ])
    

    rHand_dict = copy.deepcopy(mediaPipeHandIds.copy()) #copy.deepcopy() is necessary to make sure the dicts are independent of each other
    lHand_dict = copy.deepcopy(rHand_dict)

    for key in rHand_dict: 
        rHand_dict[key] += rHandIDstart 
        lHand_dict[key] += lHandIDstart 

    dict_of_mediaPipeSegmentIdx_dicts['mp_rHand'] = rHand_dict
    dict_of_mediaPipeSegmentIdx_dicts['mp_lHand'] = lHand_dict

    
    # #how to face --> :D <--
    # mediaPipeFaceIDs = dict()
    # faceIDStart = 67
    # #define face parts
    # mediaPipeFaceIDs['mp_jaw'] = np.arange(0,16) + faceIDStart 
    # mediaPipeFaceIDs['mp_rBrow'] = np.arange(17,21) + faceIDStart
    # mediaPipeFaceIDs['mp_lBrow'] = np.arange(22,26) + faceIDStart
    # mediaPipeFaceIDs['mp_noseRidge'] = np.arange(27,30) + faceIDStart
    # mediaPipeFaceIDs['mp_noseBot'] = np.arange(31,35) + faceIDStart
    # mediaPipeFaceIDs['mp_rEye'] = np.concatenate((np.arange(36,41), [36])) + faceIDStart
    # mediaPipeFaceIDs['mp_lEye'] = np.concatenate((np.arange(42,47), [42])) + faceIDStart    
    # mediaPipeFaceIDs['mp_upperLip'] = np.concatenate((np.arange(48,54), np.flip(np.arange(60, 64)), [48])) + faceIDStart
    # mediaPipeFaceIDs['mp_lowerLip'] = np.concatenate(([60], np.arange(64,67), np.arange(54, 59), [48], [60])) + faceIDStart
    # mediaPipeFaceIDs['mp_rPupil'] = np.array([68]) + faceIDStart
    # mediaPipeFaceIDs['mp_lPupil'] = np.array([69]) + faceIDStart #nice

    # dict_of_mediaPipeSegmentIdx_dicts['mp_face'] = mediaPipeFaceIDs
    
    return dict_of_mediaPipeSegmentIdx_dicts, dict_of_mp_skel_lineColor


if __name__ == '__main__':
    PlaySkeletonAnimation()



import bpy 
from pathlib import Path

#%% Build the User Interface panel
class VIEW3D_PT_freemocap(bpy.types.Panel): 
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "FreeMoCap"
    bl_label = "FMC_LoadSession"

    def draw(self,context):
        layout = self.layout

        #######################################################
        # UI_COLUMN_0 - The part of the panel where you specify the Session Folder        
        ui_set_session_path = layout.column(align=True)
        ui_set_session_path.prop(context.scene, 'fmc_session_path' )
        
        if context.scene.fmc_session_path : #once you set the config path            
            session_path = Path(context.scene.fmc_session_path)
            if session_path.is_dir():                #JSM NOTE - This is kinda sloppy, but I don't know how to alter the actual session path variable. I want there to be a way to enforce that the 'session_path' value points to the whole session folder, NOT a file within that folder
                ui_set_session_path.label(text=f'Session Path: {str(session_path)}') 
            else:
                ui_set_session_path.label(text=f'Session Path: {str(session_path.parent)}') 
        else: 
            ui_set_session_path.label(text='-No Session Config Loaded-')
        ui_set_session_path.label(text='_______')
        
        #######################################################
        
        ui_column_1 = self.layout.column(align=True)

        ui_column_1.operator('fmc.load_marker_empties',
            text='Load Marker Empties!',
            icon = 'EMPTY_AXIS')


        #######################################################
        
        ui_as_bones_button = self.layout.column(align=True)

        ui_as_bones_button.operator('fmc.load_raw_mocap_data_as_bones',
            text='Load raw mocap data as bones!',
            icon = 'GROUP_BONE')

        #######################################################
        
        ui_build_armature= self.layout.column(align=True)

        ui_build_armature.operator('fmc.build_armature_from_raw_mocap_data',
            text='build armature from raw mocap data',
            icon = 'ARMATURE_DATA')




        #######################################################
        ui_column_2= self.layout.column(align=True)

        ui_column_2.operator('fmc.load_nsynced_videos',
            text='Load nSynced Videos!!',
            icon = 'CAMERA_DATA')

        # properties_0 = ui_column_2.operator('fmc.load_session_data',
        #     text='Load FreeMoCap Session (Marshmallow Style!)',
        #     icon = 'VOLUME_DATA')
        # properties_0.bodyMar_size = 10
        # properties_0.faceMar_size = 10
        # properties_0.handMar_size = 10

        # properties_1 = ui_column_2.operator('fmc.load_session_data',
        #     text='Load FreeMoCap Session (Monkey Style)', 
        #     icon = 'MONKEY')
        # properties_1.meshType = 'monkey'


        #######################################################
        # UI_COLUMN_3  #The part of the panel where you click PLAY and whatnot
        ui_column_3 = self.layout.column(align=True)
        ui_column_3.prop(context.scene, 'frame_current', text='Current Frame#' )
        ui_column_3.operator('screen.animation_play', icon='PLAY')




from freemocap import fmc_runme,recordingconfig

import zipfile
from pathlib import Path
import tkinter as tk
from tkinter import Tk, Label, Button, Frame, Listbox, Entry, filedialog
import requests
import io

def DemoSetup():

    
    download_data = runDataDownloadGUI() #ask for user input on downloading
    zip_file_url = 'https://figshare.com/ndownloader/files/30913897'
    sample_session_name = 'sesh_21-07-20_165209'
    sample_session_zip = sample_session_name + '.zip'

    if  download_data: #if they want to download
        download_it = True
        #code to download stuff from online will go here
        destinationPath = runDemoGUI() 

        if destinationPath == None:
            print('Please choose where to save your sample data')
            download_it = False
            extractToPath = None
        #error seems to expect the sample session to still be a zip file even though it is unzipped below
    if download_it == True:
        currentPath = Path(destinationPath)/sample_session_name

        if 'FreeMocap_Data' in str(destinationPath):
            if currentPath.exists():
                extractToPath = None 
                print('There is already a sample data folder at: ' + str(currentPath))
                download_it = False
            #extractToPath = destinationPath/sample_session_zip
            extractToPath = destinationPath
        else:
            extractToFolder = Path(destinationPath)/recordingconfig.dataFolder #append the name of the data folder to the destination path
            extractToFolder.mkdir(exist_ok=True) #make the data folder if it doesn't current exist
            extractToPath = extractToFolder
            
            checkPath = extractToFolder/sample_session_name
            if checkPath.exists():
                print('There is already a sample data folder at: ' + str(checkPath))
                extractToPath = None 
                download_it = False
            #extractToPath = extractToFolder/sample_session_name
   

        #run the GUI to get the directory locations of the zipped folder, and where to extract it to
        #locationPath, destinationPath = runDemoGUI() 
        #zipPath = Path(locationPath)/'sesh_21-07-18_170130.zip' #append the name of the zipped sample file to the path to find it
 
        if download_it:
            print('Starting demo session download')
            r = requests.get(zip_file_url)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            z.extractall(extractToPath)
            extractedPath = Path(extractToPath)/sample_session_name
            print('Demo session downloaded to: ' + str(extractedPath))
            print('You can find the animation in the sample data folder as {}_outvid.mp4'.format(sample_session_name))
        #with zipfile.ZipFile(zipPath,'r') as zip_ref:
        #    zip_ref.extractall(extractToPath) #extract stuff to the data folder
    # else:
    #     root = tk.Tk() #bring up a window to select the data folder directory
    #     root.withdraw()
    #     destinationPath = filedialog.askdirectory( title='Select the directory where the Freemocap_data folder with the sample data is located',
    #         initialdir= Path.cwd())


    return extractToPath, sample_session_name #take the path to the directory where the data folder is located

class DemoGUI(): #GUI to pick the location and the destination directories for the sample data 
        def __init__(self, master):
            self.master = master
            
            self.destinationPath = None

            sampleDestinationText = "Select your {} folder to save the sample data into. \n If you don't have a {} folder, choose where you'd like to create one. Your sample data will be saved into it.".format(recordingconfig.dataFolder,recordingconfig.dataFolder)
            sampleDestinationLabel = Label(master, text=sampleDestinationText)
            sampleDestinationLabel.pack()

            self.destinationPathVar = tk.StringVar()
            savedDestinationLabel = Label(master,textvariable=self.destinationPathVar)
            savedDestinationLabel.pack()

            sampleDestinationButton = Button(master,text = 'Select the sample data destination',command = self.openDestinationDialog)
            sampleDestinationButton.pack()

            submitButton = Button(master,text = 'Submit',command = self.submit)
            submitButton.pack()


        def openDestinationDialog(self):
     
            self.destinationPath = filedialog.askdirectory(
            title='Select the directory where the freemocap data is located or should be created',
            initialdir= Path.cwd())
            f = 2

            self.destinationPathVar.set(self.destinationPath)
        
        def submit(self):
            self.master.destroy()


        def getLocationPath(self):
            return self.locationPath
        
        def getDestinationPath(self):
            return self.destinationPath

class dataDownloadGUI(): #GUI to ask whether users want to download data
    def __init__(self, master):
        self.master = master
    
        dataText = "Would you like to download sample data?"
        dataLable = Label(master, text=dataText)
        dataLable.pack(side="top")

        yesButton =  Button(master,text = 'Yes',command = self.yes)
        yesButton.pack()

        noButton =  Button(master,text = 'No',command = self.no)
        noButton.pack()

    def yes(self):
        self.download_sample_data = True
        self.master.destroy()
        

    def no(self):
        self.download_sample_data = False
        self.master.destroy()


def runDemoGUI():
    root = tk.Tk()
    pathchoices = DemoGUI(root)
    root.mainloop()
    destinationPath = pathchoices.destinationPath

    return destinationPath

def runDataDownloadGUI():
    root = tk.Tk()
    download_options = dataDownloadGUI(root)
    root.mainloop()

    return download_options.download_sample_data





# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

bl_info = {
    "name" : "FreeMoCap",
    "author" : "Jonathan Samir Matthis",
    "description" : "",
    "blender" : (2, 80, 0),
    "version" : (0, 0, 1),
    "location" : "",
    "warning" : "",
    "category" : "fmc"
}

import bpy
from . import auto_load

auto_load.init()

def register():
    
    #define 'custom properties' - https://youtu.be/9fuFDHR-UkE 

    bpy.types.Scene.fmc_session_path = bpy.props.StringProperty(
        name = 'FMC Session Path',
        subtype= 'FILE_PATH'
    )

    auto_load.register()
    print('registering FreeMoCap Addon')

def unregister():
    auto_load.unregister()



import threading
import cv2
import time
import pickle
import os
import platform

class CamRecordingThread(threading.Thread):
    def __init__(
        self, session, camID, camInput, videoName, rawVidPath, beginTime, parameterDictionary
    ):
        threading.Thread.__init__(self)
        self.camID = camID
        self.camInput = camInput
        self.videoName = videoName
        self.rawVidPath = rawVidPath
        self.beginTime = beginTime
        self.parameterDictionary = parameterDictionary
        self.session = session

    def run(self):
        print("Starting " + self.camID)
        self.timeStamps = CamRecording(
            self.session,
            self.camID,
            self.camInput,
            self.videoName,
            self.rawVidPath,
            self.beginTime,
            self.parameterDictionary,
        )

    def getStamps(self):
        return self.timeStamps


# the recording function that each threaded camera object runs
def CamRecording(
    session, camID, camInput, videoName, rawVidPath, beginTime, parameterDictionary
):
    """
    Runs the recording process for each threaded camera instance. Saves a video to the RawVideos folder.
    Saves a timestamp for each frame into a pickle file. Checks for whether the global variable 'flag' 
    has been set to True to end the recording (the first camera to end sets the flag to True, which
    means the rest of the cameras will quit in short order)
    """
    # the flag is triggered when the user shuts down one webcam to shut down the rest.
    # normally I'd try to avoid global variables, but in this case it's
    # necessary, since each webcam runs as it's own object.
    global flag
    flag = False
    camWindowName = "RECORDING - " + str(camID)+' - Press ESC to exit'
    cv2.namedWindow(camWindowName)  # name the preview window for the camera its showing

    if platform.system() == 'Windows':
        cam = cv2.VideoCapture(camInput, cv2.CAP_DSHOW)
    else:
        cam = cv2.VideoCapture(camInput, cv2.CAP_ANY)


    # if not cam.isOpened():
    #         raise RuntimeError('No camera found at input '+ str(camID))
    # pulling out all the dictionary paramters
    exposure = parameterDictionary.get("exposure")
    resWidth = parameterDictionary.get("resWidth")
    resHeight = parameterDictionary.get("resHeight")
    framerate = parameterDictionary.get("framerate")
    codec = parameterDictionary.get("codec")


  

    cam.set(cv2.CAP_PROP_FRAME_WIDTH, resWidth)
    cam.set(cv2.CAP_PROP_FRAME_HEIGHT, resHeight)
    cam.set(cv2.CAP_PROP_EXPOSURE, exposure)
    cam.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')) 
    fourcc = cv2.VideoWriter_fourcc(*codec)
    # rawPath = filepath/'RawVideos' #creating a RawVideos folder
    # rawPath.mkdir(parents = True, exist_ok = True)
    width = cam.get(cv2.CAP_PROP_FRAME_WIDTH)
    height = cam.get(cv2.CAP_PROP_FRAME_HEIGHT)
    print("width:", width, "height:", height)
    saveRawVidPath = str( rawVidPath / videoName)  # create a save path for each video to the RawVideos folders
    
    out = cv2.VideoWriter(saveRawVidPath, fourcc, framerate, (resWidth, resHeight))
    timeStamps = []  # holds the timestamps

    if cam.isOpened():
        success, frame = cam.read()
    else:
        success = False

    while ( success):  # while the camera is opened, record the data until the escape button is hit
        if flag:  # when the flag is triggered, stop recording and dump the data
            with open(session.rawVidPath/camID, "wb") as f:
                pickle.dump(timeStamps, f)
            break
        success, frame = cam.read()

        cv2.imshow(camWindowName, frame)
        frame_sized = cv2.resize(frame, (resWidth, resHeight))
        frame_sized = frame
        out.write(frame)
        timeStamps.append(time.time() - beginTime)  # add each timestamp to the list

        key = cv2.waitKey(20)
        if key == 27:  # exit on ESC
            flag = True  # set flag to true to shut down all other webcams
            with open(session.rawVidPath/camID, "wb") as f:
                pickle.dump(timeStamps, f)  # dump the data
            break
    cv2.destroyWindow(camWindowName)
    return timeStamps


# this is how we sync our time frames, based on our recorded timestamps




from scipy.ndimage.measurements import center_of_mass
from freemocap.fmc_startup import startup, startupGUI
from freemocap.webcam import camera_settings, timesync

from pathlib import Path
import os
import subprocess
import time 
from aniposelib.boards import CharucoBoard

import numpy as np
from scipy.signal import savgol_filter

from ruamel.yaml import YAML
import cv2

#Rich stuff
from rich import print
from rich.console import Console
console = Console()
from rich.markdown import Markdown
from rich.traceback import install
install(show_locals=False)
from rich import inspect
from rich.padding import Padding


from freemocap import (
    recordingconfig,
    runcams,
    calibrate,
    fmc_mediapipe,
    fmc_openpose,
    fmc_deeplabcut,
    reconstruct3D,
    play_skeleton_animation,
    session,
)



thisStage = 0 #global

# TODO: Replace the below functions with the RunMe options.
def RunMe(sessionID=None,
        stage=1,
        useOpenPose=False, 
        runOpenPose = True, 
        useMediaPipe=True,
        runMediaPipe=True,
        useDLC=False,
        dlcConfigPath=None,
        debug=False,
        setDataPath = False,
        userDataPath = None,
        recordVid = True,
        showAnimation = True,
        reconstructionConfidenceThreshold = .7,
        charucoSquareSize = 36,#mm - ~the size of the squares when printed on 8.5x11" paper based on parameters in ReadMe.md
        calVideoFrameLength = .5,
        startFrame = 0,
        useBlender = False,
        resetBlenderExe = False,
        ):
    """ 
    Starts the freemocap pipeline based on either user-input values, or default values. Creates a new session class instance (called sesh)
    based on the specified inputs. Checks for previous user preferences and choices if they exist, or will prompt the user for new choices
    if they don't. Runs the initialization for the system and runs each stage of the pipeline.
    """ 
    
    welcome_md = Markdown("""# Welcome to FreeMoCap ✨💀✨ """)
    console.print(welcome_md)
    
    sesh = session.Session()

    sesh.sessionID = sessionID
    sesh.useOpenPose= useOpenPose
    sesh.useMediaPipe = useMediaPipe
    sesh.useDLC= useDLC
    sesh.debug = debug
    sesh.setDataPath = setDataPath
    sesh.userDataPath = userDataPath
    sesh.dataFolderName = recordingconfig.dataFolder
    sesh.startFrame = startFrame

    # %% Startup 
    startup.get_user_preferences(sesh,stage)

    if sesh.useDLC and stage<5:
         import deeplabcut as dlc
         dlc_config_paths = startup.get_dlc_paths(session)

    if stage > 1:
        startup.get_data_folder_path(sesh)
    
        if sesh.sessionID == None:    
            subfolders = [f.path for f in os.scandir(sesh.dataFolderPath) if f.is_dir()]  # copy-pasta from who knows where
            sesh.sessionID = Path(subfolders[-1]).stem  # grab the name of the last folder in the list of subfolders
        
        console.rule()
        print('Running ' + str(sesh.sessionID) + ' from ' + str(sesh.dataFolderPath))
        console.rule()
    if useBlender == True:
        here = Path(__file__).parent
        subprocessPath = here/'fmc_blender.py'
        blenderPath = startup.get_blender_path(sesh,resetBlenderExe)

 
    board = CharucoBoard(7, 5,
                        #square_length=1, # here, in mm but any unit works (JSM NOTE - just using '1' so resulting units will be in 'charuco squarelenghts`)
                        #marker_length=.8,
                        #  square_length = 121, #big boi charuco
                        #  marker_length = 98,
                        square_length = charucoSquareSize,#mm
                        marker_length = charucoSquareSize*.8,#mm
                        marker_bits=4, dict_size=250)
    sesh.board = board

   

    # %% Initialization
    if stage == 1:
        camera_settings.initialize(sesh,stage,board)
    elif stage ==2:
        timesync.time_sync_initialize(sesh)
    else:
        sesh.initialize(stage)

    # %% Stage One
    
    if stage <= 1:
        thisStage=1
        console.rule(style="color({})".format(14))    
        console.rule('Starting Video Recordings'.upper(), style="color({})".format(14))   
        console.rule(style="color({})".format(14))    

        runcams.RecordCams(sesh, sesh.cam_inputs, sesh.parameterDictionary, sesh.rotationInputs)
        sesh.save_session()
    else:
        print('Skipping Video Recording')


    # %% Stage Two
    if stage <= 2:
        thisStage=2
        console.rule(style="color({})".format(thisStage))        
        console.rule('Synchronizing Recorded Videos'.upper(),style="color({})".format(thisStage))    
        console.rule(style="color({})".format(thisStage))    
        
        runcams.SyncCams(sesh, sesh.timeStampData,sesh.numCamRange,sesh.vidNames,sesh.camIDs)
        sesh.save_session()
    else:
        print('Skipping Video Syncing')

    # %% Stage Three
    if stage <= 3:
        thisStage=3
        console.rule(style="color({})".format(thisStage))    
        console.rule('Starting Capture Volume Calibration'.upper(),style="color({})".format(thisStage))    
        console.rule(style="color({})".format(thisStage))  
        console.print(Padding('Using Anipose to calculate 6 degree-of-freedom position (and distortion coeffs) of each camera based on detected charuco boards. This information is used to create a Camera Projection Matrix for each camera, which is later used in the 3d reconstruction stage', (1,4)), overflow="fold", justify='center',style="color({})".format(thisStage))
        console.rule('See https://anipose.org for details', style="color({})".format(thisStage))  
        console.rule(style="color({})".format(thisStage))    

        if sesh.numFrames is None:
            a_sync_vid_path = list(sesh.syncedVidPath.glob('*.mp4'))
            temp_cap =   cv2.VideoCapture(str(a_sync_vid_path[0]))
            sesh.numFrames = temp_cap.get(cv2.CAP_PROP_FRAME_COUNT)
            temp_cap.release()

        sesh.cgroup, sesh.mean_charuco_fr_mar_xyz = calibrate.CalibrateCaptureVolume(sesh,board, calVideoFrameLength)

        ##this is supposed to cycle through tyhe videos with different windows to try to get Anipose to work. I can't get the dang thing working because something weird happens where a Thread will get spawned in one of the inner functions (maybe related to tqdm?) and that iteration will jump out of the try/except 
        # try:
        #     sesh.cgroup, sesh.mean_charuco_fr_mar_xyz = calibrate.CalibrateCaptureVolume(sesh,board, calVideoFrameLength)
        #     # anipose_success = True
        #     anipose_success = False
        # except:
        #     console.print_exception()
        #     console.print('[bold red] - Anipose Calibration failed with user-provided (or default) `calVideoFrameLength` value! Trying again with other parts of the videos')
        #     anipose_success = False

        # if not anipose_success:
        #     if calVideoFrameLength ==.25:
        #         cal_video_frame_range = [round(sesh.numFrames*.25), round(sesh.numFrames*.5)]
        #     else:
        #         cal_video_frame_range = [0, round(sesh.numFrames*.25)]
            
        #     for anipose_iter in range(4):
                
        #         console.rule('Anipose Failed - Reprocessing - Iteration #{}'.format(anipose_iter), style="color({})".format(thisStage))
        #         console.rule('Trying again with frame range {} - {}'.format(cal_video_frame_range[0], cal_video_frame_range[1]), style="color({})".format(thisStage))

        #         try:
        #             sesh.cgroup, sesh.mean_charuco_fr_mar_xyz = calibrate.CalibrateCaptureVolume(sesh,board, cal_video_frame_range)
        #             break
        #         except:
        #             console.print_exception()
        #             cal_video_frame_range = [cal_video_frame_range[0]+round(sesh.numFrames*.25), cal_video_frame_range[0]+round(sesh.numFrames*.5)]
                
        #         if cal_video_frame_range[1] > sesh.numFrames:
        #             console.print('[bold red] -Sorry, we couldn\'t get Anipose calibration to complete sucessfully. Are you using a Charuco board made with the parameters described in the ReadMe (here\s a sample png -https://github.com/jonmatthis/freemocap/blob/main/charuco_board_image.png ). Is the board clearly visible to each camera? Is there glare on it from from any of the camera\'s perspective? Is it too far away from the cameras? Is your `exposure` set low enough that the black squares are black (not grey)?')

                


        print('Anipose Calibration Successful!')
    else:
        print('Skipping Calibration')

    # %% Stage Four
    if stage <= 4:
        thisStage=4
        console.rule(style="color({})".format(thisStage))    
        console.rule('Starting 2D Point Trackers'.upper(),style="color({})".format(thisStage))  
        stage4_msg ='This step implements various  computer vision that track the skeleton (and other objects) in the 2d videos, to produce the data that will be combined with the `camera projection matrices` from the calibration stage to produce the estimates of 3d movement. \n \n Each algorithm is different, but most involve using [bold magenta] convolutional neural networks [/bold magenta] trained from labeled videos to produce a 2d probability map of the likelihood that the tracked bodypart/object/feature (e.g. \'LeftElbow\') is in a given location. \n \n The peak of that distrubtion on each frame is recorded as the pixel-location of that item on that frame (e.g. \'LeftElbow(pixel-x, pixel-y, confidence\') where the a confidence value proportional to the underlying probability distribution (i.e. tall peaks in the probablitiy distribution indicate high confidence that the LeftElbow actually is at this pixel-x, pixel-y location) \n \nThis part is crazy future tech sci fi stuff. Seriously unbelievable this kind of thing is possible ✨'
        console.print(Padding(stage4_msg, (1,4)), overflow="fold", justify='center',style="color({})".format(thisStage))
        console.rule(style="color({})".format(thisStage))      
        

        if sesh.useMediaPipe:
            console.rule(style="color({})".format(thisStage))    
            console.rule('Running MediaPipe skeleton tracker - https://google.github.io/mediapipe', style="color({})".format(thisStage))    
            console.rule(style="color({})".format(thisStage))    
            

            if runMediaPipe:
                fmc_mediapipe.runMediaPipe(sesh)
                sesh.mediaPipeData_nCams_nFrames_nImgPts_XYC = fmc_mediapipe.parseMediaPipe(sesh)
                sesh.mediaPipeSkel_fr_mar_xyz, sesh.mediaPipeSkel_reprojErr = reconstruct3D.reconstruct3D(sesh,sesh.mediaPipeData_nCams_nFrames_nImgPts_XYC, confidenceThreshold=reconstructionConfidenceThreshold)
                np.save(sesh.dataArrayPath/'mediaPipeSkel_3d.npy', sesh.mediaPipeSkel_fr_mar_xyz) #save data to npy
                np.save(sesh.dataArrayPath/'mediaPipeSkel_reprojErr.npy', sesh.mediaPipeSkel_reprojErr) #save data to npy            


                smoothWinLength = 5
                smoothOrder = 3
                for dim in range(sesh.mediaPipeSkel_fr_mar_xyz.shape[2]):
                    for mm in range(sesh.mediaPipeSkel_fr_mar_xyz.shape[1]):
                        sesh.mediaPipeSkel_fr_mar_xyz[:,mm,dim] = savgol_filter(sesh.mediaPipeSkel_fr_mar_xyz[:,mm,dim], smoothWinLength, smoothOrder)

                np.save(sesh.dataArrayPath/'mediaPipeSkel_3d_smoothed.npy', sesh.mediaPipeSkel_fr_mar_xyz) #save data to npy

        sesh.save_session()

        if sesh.useOpenPose:
            console.rule(style="color({})".format(thisStage))    
            console.rule('Running OpenPose skeleton tracker - https://github.com/CMU-Perceptual-Computing-Lab/openpose', style="color({})".format(thisStage))    
            console.rule(style="color({})".format(thisStage))
                

            fmc_openpose.runOpenPose(sesh, runOpenPose=runOpenPose)
            sesh.openPoseData_nCams_nFrames_nImgPts_XYC = fmc_openpose.parseOpenPose(sesh)
            sesh.openPoseSkel_fr_mar_xyz, sesh.openPoseSkel_reprojErr = reconstruct3D.reconstruct3D(sesh,sesh.openPoseData_nCams_nFrames_nImgPts_XYC, confidenceThreshold=reconstructionConfidenceThreshold)
            np.save(sesh.dataArrayPath/'openPoseSkel_3d.npy', sesh.openPoseSkel_fr_mar_xyz) #save data to npy
            np.save(sesh.dataArrayPath/'openPoseSkel_reprojErr.npy', sesh.openPoseSkel_reprojErr) #save data to npy
            
            smoothWinLength = 5
            smoothOrder = 3
            for dim in range(sesh.openPoseSkel_fr_mar_xyz.shape[2]):
                for mm in range(sesh.openPoseSkel_fr_mar_xyz.shape[1]):
                    sesh.openPoseSkel_fr_mar_xyz[:,mm,dim] = savgol_filter(sesh.openPoseSkel_fr_mar_xyz[:,mm,dim], smoothWinLength, smoothOrder)

            np.save(sesh.dataArrayPath/'openPoseSkel_3d_smoothed.npy', sesh.openPoseSkel_fr_mar_xyz) #save data to npy
                
        sesh.save_session()
        sesh.syncedVidList = []

        if sesh.useDLC:
            
            console.rule(style="color({})".format(thisStage))    
            console.rule('Running DeepLabCut :mouse: - https://deeplabcut.org', style="color({})".format(thisStage))    
            console.rule(style="color({})".format(thisStage)) 
               

            for vid in sesh.syncedVidPath.glob('*.mp4'):
                sesh.syncedVidList.append(str(vid))
            
            for config_path in dlc_config_paths:
                dlc.analyze_videos(config_path,sesh.syncedVidList, destfolder= sesh.dlcDataPath, save_as_csv=True) 
                sesh.dlcData_nCams_nFrames_nImgPts_XYC = fmc_deeplabcut.parseDeepLabCut(sesh, config_path)
                sesh.dlc_fr_mar_xyz, sesh.dlc_reprojErr = reconstruct3D.reconstruct3D(sesh,sesh.dlcData_nCams_nFrames_nImgPts_XYC, confidenceThreshold=reconstructionConfidenceThreshold)
                np.save(sesh.dataArrayPath/'deepLabCut_3d.npy', sesh.dlc_fr_mar_xyz) #save data to npy
                np.save(sesh.dataArrayPath/'deepLabCut_reprojErr.npy', sesh.dlc_reprojErr) #save data to npy
        sesh.save_session()
    else:

        print('Skipping 2d point tracking')


    # %% Stage 5 - Use Blender to create output data files
    if stage <=5:
        try:
            if useBlender == True:
                thisStage=5
                console.rule(style="color({})".format(thisStage))    
                console.rule('Exporting Files...'.upper(), style="color({})".format(thisStage))    
                console.rule('Hijacking Blender\'s file format converters to export FreeMoCap data as various file format (.blend, .usd, .gltf, .fbx)', style="color({})".format(thisStage))    
                console.rule(style="color({})".format(thisStage))    
                


                #blenderPath = Path('C:\Program Files\Blender Foundation\Blender 2.93')
                #os.chdir(blenderExePath)
            output = subprocess.run([str(blenderPath), "--background", "--python", "freemocap/fmc_blender.py", "--", str(sesh.dataArrayPath/'mediaPipeSkel_3d_smoothed.npy'), str(sesh.dataArrayPath), sesh.sessionID], capture_output=True, text=True, check=True)
            print(output)        
        except:
            console.print_exception()
            

    # %% Stage 6 - Make  Animation
    if stage <= 6:
        thisStage=6
        console.rule(style="color({})".format(thisStage))    
        console.rule('Creating the Skreleton animation!'.upper(),style="color({})".format(thisStage))    
        console.print('The video creation is very slow. All of the animation making code is crazy slow, tbh. Sorry about that, future iterations will be better lol :sweat_smile:',overflow="fold", justify='center',style="color({})".format(thisStage))    
        console.rule(style="color({})".format(thisStage))    
        
        

        play_skeleton_animation.PlaySkeletonAnimation(
                                sesh,
                                startFrame=sesh.startFrame,
                                azimuth=-90,
                                elevation=-81,
                                useOpenPose=useOpenPose,
                                useMediaPipe=useMediaPipe,
                                useDLC=useDLC,
                                recordVid = recordVid,
                                showAnimation=showAnimation,
                                )
        console.rule(style="color({})".format(thisStage))    
    else:
        print('Skipping Skeleton Plotting')


    console.rule(style="color({})".format(13))      
    console.rule('All Done!'.upper(), style="color({})".format(13))    
    console.rule(style="color({})".format(13))    
    console.rule('Session Data folder is at: ', style="color({})".format(13))    
    console.rule(str(sesh.sessionPath), style="color({})".format(13))    
    console.rule(style="color({})".format(13))    
    console.rule(style="color({})".format(10))    
    console.rule("Thank you for supporting the FreeMoCap Project", style="color({})".format(10))    
    console.rule(style="color({})".format(10))    
    console.rule(style="color({})".format(13))    
    console.print('~✨💀✨~',justify="center") 
    console.print('❤️', justify="center") 
    console.rule(style="color({})".format(13))    
    console.rule(style="color({})".format(14))    




#this is copied more or less straight from jallex's `freemocap/fmc_blender` script. It needs some work, but it does do the job!

import bpy 
import numpy as np 
from pathlib import Path
import math
from mathutils import Vector

class FMC_OT_build_armature_from_raw_mocap_data(bpy.types.Operator): #setting the type as "FMC" for "FreeMoCap"
    """ Load marker data via FMC yaml config and place empties in Blender scene """
    bl_idname = "fmc.build_armature_from_raw_mocap_data"
    bl_label = "Build an armature from raw mocap data (in development)"
    bl_options = {'REGISTER', 'UNDO'}

  
    def execute(self, context):
        session_path = Path(context.scene.fmc_session_path)        
         
        if not session_path.is_dir():
            session_path = session_path.parent
        
        input_npy = session_path / 'DataArrays' / 'mediaPipeSkel_3d.npy'
        ##########
        #### Begin Copy for fmc_blender 
        ##########
        
        #the path of the npy build data file 
        build_data = Path(__file__).parent.resolve() / "build_data.npy"

        #3D array holding [[[x, y, z], [x, y, z]], [[x, y, z], [x, y, z]]] 
        arr = np.load(input_npy)

        # Saved clean frame where all points are visible and not NANs 
        markers_list = np.load(build_data)

        #a list containing empty objects
        order_of_markers = []

        #frame rate 
        bpy.context.scene.render.fps = 30

        #dictionary corresponding to marker numbers and body parts
        body_dict = {
            0:"head1", 
            11:"shoulder_L", 
            12:"shoulder_R", 
            13:"elbow_L", 
            14:"elbow_R", 
            23:"hip_L",
            24:"hip_R", 
            25:"knee_L", 
            26:"knee_R",
            27:"ankle_L",
            28:"ankle_R",
            29:"heel_L", 
            30:"heel_R",
            31:"toe_L",
            32:"toe_R" }

        #-----------------------------------------------------------------------------------   

        # set project unit 
        bpy.context.scene.unit_settings.length_unit = 'METERS'
        #iterate through arr and create an empty object at that location for each element
        for index, col in enumerate(markers_list):
            # parse string float value into floats, create Vector, set position to Vector

            if math.isnan(col[0]):
                col[0] = 0.0
            if math.isnan(col[1]):
                col[1] = 0.0
            if math.isnan(col[2]):
                col[2] = 0.0
            coord = Vector(((float(col[0])*0.001), (float(col[2])*0.001), (float(col[1]))* -0.001))
                
            #empties
            bpy.ops.object.add(type='EMPTY', location=coord)  
            mt = bpy.context.active_object  
            mt.name = "mt_" + str(index)
            if index in body_dict.keys():
                mt.name += "_" + str(body_dict[index])
            order_of_markers.append(mt)
            #link empty to scene
            bpy.context.scene.collection.objects.link( mt )
            #set location 
            mt.location = coord
            #set the display size of the empty
            mt.empty_display_size = 0.02
            
        #--------------------------------------------------------------
        #Virtual Markers!

        # marker relationships:

        # - "weight": the weighted average of multiple markers. the virtual_markers[x] contains the 
        #list of markers that affect this virtual one. the weights[x] contains their corresponding weights in order.

        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        #Create virtual marker where parameters are the name of the marker, the markers that affect its position, and each of their weights.
        def create_marker_weight(name, markers, weighted):
            center = Vector((0, 0, 0))
            weight_iter = 0
            for x in markers:
                center += x.location*weighted[weight_iter]
                weight_iter += 1
            coord = Vector((float(center[0]), float(center[1]), float(center[2])))
            bpy.ops.object.add(type='EMPTY', location=coord)
            mt = bpy.context.active_object  
            mt.name = name
            bpy.context.scene.collection.objects.link( mt )
            mt.location = coord
            mt.empty_display_size = 0.02
            virtual_markers.append(mt)

        #Keeping track of virtual marker info using arrays, where each marker is an index in each array
        v_relationship = []
        virtual_markers = []
        surrounding_markers = []
        weights = []

        #updates data and creates virtual marker
        def update_virtual_data(relationship, surrounding, vweights, vname):
            v_relationship.append(relationship)
            surrounding_markers.append(surrounding)
            weights.append(vweights)
            create_marker_weight(vname, surrounding, vweights)
                
        #-----------------------------------------------------------------
        #Define relationships and create virtual markers

        #Neck Base: Halfway between order_of_markers[12] shoulder_L and order_of_markers[12] shoulder_R
        l0 = [order_of_markers[12], order_of_markers[11]]
        w0 = [0.5, 0.5]
        update_virtual_data("weight", l0, w0, "m_neck")

        #Waist Base: Halfway between order_of_markers[24] and order_of_markers[23] 
        l0 = [order_of_markers[24], order_of_markers[23]]
        w0 = [0.5, 0.5]
        update_virtual_data("weight", l0, w0, "m_neck")

        #Update the location of virtual markers on each frame
        def update_virtual_marker(index):
            if(v_relationship[index] == "weight"):
                center = Vector((0, 0, 0))
                weight_iter = 0
                for x in surrounding_markers[index]:
                    center += x.location*weights[index][weight_iter]
                    weight_iter += 1
                coord = Vector((float(center[0]), float(center[1]), float(center[2])))
            virtual_markers[index].location = coord
            
        #-----------------------------------------------------------------------------------
        # Bones! 
            
        #adds child bone given corresponding parent and empty
        #bone tail will appear at the location of empty
        def add_child_bone(bone_name, empty1, empty2):
            #Set armature selected
            armature_data.select_set(state=True)
            #Set edit mode
            bpy.ops.object.mode_set(mode='EDIT', toggle=False)
            #Create a new bone
            new_bone = armature_data.data.edit_bones.new(bone_name)
            #Set bone's size
            new_bone.head = (0,0,0)
            new_bone.tail = (0,0.1,0)
            #Set bone's location to wheel
            new_bone.matrix = empty2.matrix_world
            #set location of bone head
            new_bone.head =  empty1.location
            #set location of bone tail
            new_bone.tail = empty2.location
            return new_bone

        #Create armature object
        armature = bpy.data.armatures.new('Armature')
        armature_object = bpy.data.objects.new('Armature', armature)
        #Link armature object to our scene
        bpy.context.collection.objects.link(armature_object)
        #Make armature variable
        armature_data = bpy.data.objects[armature_object.name]
        #Set armature active
        bpy.context.view_layer.objects.active = armature_data
        #Set armature selected
        armature_data.select_set(state=True)
        #Set edit mode
        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        #Set bones in front and show axis
        armature_data.show_in_front = True
        #True to show axis orientation of bones and false to hide it
        armature_data.data.show_axes = False

        #get armature object
        def get_armature():
            for ob in bpy.data.objects:
                if ob.type == 'ARMATURE':
                    armature = ob
                    break
            return armature

        armature = get_armature()

        #a list of tuples where each element represents the info for one bone, and 
        #the tuple contains the bone name, the marker of the head, and the marker of the tail
        list_of_bones_order = [('upper_arm_L', order_of_markers[11], order_of_markers[13]),
                ('upper_arm_R', order_of_markers[12], order_of_markers[14]),
                ('lower_arm_L', order_of_markers[13], order_of_markers[15]),
                ('lower_arm_R', order_of_markers[14], order_of_markers[16]),
                ('upper_leg_L', order_of_markers[23], order_of_markers[25]),
                ('upper_leg_R', order_of_markers[24], order_of_markers[26]),
                ('lower_leg_L', order_of_markers[25], order_of_markers[27]),
                ('lower_leg_R', order_of_markers[26], order_of_markers[28]),
                ('heel_L', order_of_markers[27], order_of_markers[29]),
                ('heel_R', order_of_markers[28], order_of_markers[30]),
                ('foot_L', order_of_markers[29], order_of_markers[31]),
                ('foot_R', order_of_markers[30], order_of_markers[32]),
                ('neck', order_of_markers[0], virtual_markers[0]),
                ('eye_L', order_of_markers[0], order_of_markers[2]),
                ('ear_L', order_of_markers[2], order_of_markers[7]),
                ('eye_R', order_of_markers[0], order_of_markers[5]),
                ('ear_R', order_of_markers[5], order_of_markers[8]),
                ('shoulder_L', virtual_markers[0], order_of_markers[11]),
                ('shoulder_R', virtual_markers[0], order_of_markers[12]),
                ('torso', virtual_markers[1], virtual_markers[0]),
                ('hip_L', virtual_markers[1], order_of_markers[23]),
                ('hip_R', virtual_markers[1], order_of_markers[24]),
                ]
                
        #based on marker # from order_of_markers array add bones for hands:
        left_hand_offset = 54
        left_hand = [('handL0', order_of_markers[0+left_hand_offset], order_of_markers[1+left_hand_offset]),
            ('handL1', order_of_markers[1+left_hand_offset], order_of_markers[2+left_hand_offset]),
            ('handL2', order_of_markers[2+left_hand_offset], order_of_markers[3+left_hand_offset]),
            ('handL3', order_of_markers[3+left_hand_offset], order_of_markers[4+left_hand_offset]),
            ('handL5', order_of_markers[1+left_hand_offset], order_of_markers[5+left_hand_offset]),
            ('handL6', order_of_markers[5+left_hand_offset], order_of_markers[6+left_hand_offset]),
            ('handL7', order_of_markers[6+left_hand_offset], order_of_markers[7+left_hand_offset]),
            ('handL8', order_of_markers[7+left_hand_offset], order_of_markers[8+left_hand_offset]),
            ('handL9', order_of_markers[5+left_hand_offset], order_of_markers[9+left_hand_offset]),
            ('handL10', order_of_markers[9+left_hand_offset], order_of_markers[10+left_hand_offset]),
            ('handL11', order_of_markers[10+left_hand_offset], order_of_markers[11+left_hand_offset]),
            ('handL12', order_of_markers[11+left_hand_offset], order_of_markers[12+left_hand_offset]),
            ('handL13', order_of_markers[9+left_hand_offset], order_of_markers[13+left_hand_offset]),
            ('handL14', order_of_markers[13+left_hand_offset], order_of_markers[14+left_hand_offset]),
            ('handL15', order_of_markers[14+left_hand_offset], order_of_markers[15+left_hand_offset]),
            ('handL16', order_of_markers[15+left_hand_offset], order_of_markers[16+left_hand_offset]),
            ('handL17', order_of_markers[13+left_hand_offset], order_of_markers[0+left_hand_offset]),
            ('handL18', order_of_markers[0+left_hand_offset], order_of_markers[17+left_hand_offset]),
            ('handL19', order_of_markers[17+left_hand_offset], order_of_markers[18+left_hand_offset]),
            ('handL20', order_of_markers[18+left_hand_offset], order_of_markers[19+left_hand_offset]),
            ('wristL', order_of_markers[15], order_of_markers[0+left_hand_offset]),
            ('handL21', order_of_markers[19+left_hand_offset], order_of_markers[20+left_hand_offset]),]
            
        right_hand_offset = 33

        right_hand = [('handR0', order_of_markers[18], order_of_markers[1+right_hand_offset]),
            ('handR1', order_of_markers[1+right_hand_offset], order_of_markers[2+right_hand_offset]),
            ('handR2', order_of_markers[2+right_hand_offset], order_of_markers[3+right_hand_offset]),
            ('handR3', order_of_markers[3+right_hand_offset], order_of_markers[4+right_hand_offset]),
            ('handR5', order_of_markers[1+right_hand_offset], order_of_markers[5+right_hand_offset]),
            ('handR6', order_of_markers[5+right_hand_offset], order_of_markers[6+right_hand_offset]),
            ('handR7', order_of_markers[6+right_hand_offset], order_of_markers[7+right_hand_offset]),
            ('handR8', order_of_markers[7+right_hand_offset], order_of_markers[8+right_hand_offset]),
            ('handR9', order_of_markers[5+right_hand_offset], order_of_markers[9+right_hand_offset]),
            ('handR10', order_of_markers[9+right_hand_offset], order_of_markers[10+right_hand_offset]),
            ('handR11', order_of_markers[10+right_hand_offset], order_of_markers[11+right_hand_offset]),
            ('handR12', order_of_markers[11+right_hand_offset], order_of_markers[12+right_hand_offset]),
            ('handR13', order_of_markers[9+right_hand_offset], order_of_markers[13+right_hand_offset]),
            ('handR14', order_of_markers[13+right_hand_offset], order_of_markers[14+right_hand_offset]),
            ('handR15', order_of_markers[14+right_hand_offset], order_of_markers[15+right_hand_offset]),
            ('handR16', order_of_markers[15+right_hand_offset], order_of_markers[16+right_hand_offset]),
            ('handR17', order_of_markers[18], order_of_markers[13+right_hand_offset]),
            ('handR18', order_of_markers[18], order_of_markers[17+right_hand_offset]),
            ('handR19', order_of_markers[17+right_hand_offset], order_of_markers[18+right_hand_offset]),
            ('handR20', order_of_markers[18+right_hand_offset], order_of_markers[19+right_hand_offset]),
            ('wristR', order_of_markers[16], order_of_markers[18]),
            ('handR21', order_of_markers[19+right_hand_offset], order_of_markers[20+right_hand_offset]),]

        #helper to create armature from list of tuples
        def tuple_to_armature(bones):
            for bone_name, bone_head, bone_tail in bones:
                add_child_bone(bone_name, bone_head, bone_tail)
                
        #Set armature selected
        armature_data.select_set(state=True)
        #Set edit mode
        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                
        #create all bones for skeleton body and hands
        tuple_to_armature(list_of_bones_order)
        tuple_to_armature(right_hand)
        tuple_to_armature(left_hand)

        #parent heads and tails to empties
        #use bone constraints 
        def parent_to_empties(bone_name, head, tail):
            #enter pose mode
            bpy.ops.object.posemode_toggle()
            marker = armature.data.bones[bone_name]
            #Set marker selected
            marker.select = True
            #Set marker active
            bpy.context.object.data.bones.active = marker
            bone = bpy.context.object.pose.bones[bone_name]
            #Copy Location Pose constraint: makes the bone's head follow the given empty
            bpy.ops.pose.constraint_add(type='COPY_LOCATION')
            bone.constraints["Copy Location"].target = head
            #Stretch To Pose constraint: makes the bone's tail follow the given empty
            #stretches the bones to reach the tail to that empty so head location is not affected
            bpy.ops.pose.constraint_add(type='STRETCH_TO')
            bone.constraints["Stretch To"].target = tail
            #exit pose mode
            bpy.ops.object.posemode_toggle()
            
        #set parents of heads and tails for each bone 
        def tuple_to_parented(bones):
            for bone_name, bone_head, bone_tail in bones:
                parent_to_empties(bone_name, bone_head, bone_tail)

        tuple_to_parented(list_of_bones_order)
        tuple_to_parented(right_hand)
        tuple_to_parented(left_hand)

        #-----------------------------------------------------------------------------------
        # Animate! 
        #find number of frames in animation
        num_frames = len(arr)

        #change start frame of animation
        bpy.context.scene.frame_start = 1
        #change end frame of animation
        bpy.context.scene.frame_end = num_frames - 1

        #create a new handler to change empty positions every frame
        def my_handler(scene): 
            bpy.ops.object.mode_set(mode='OBJECT')
            #keep track of current_marker
            current_marker = 0
            #find the current frame number
            frame = scene.frame_current
            #get the list of marker points from the current frame
            markers_list = arr[frame]
            #iterate through list of markers in this frame
            for col in markers_list:
                frame = scene.frame_current
                if math.isnan(col[0]):
                    col[0] = 0.0
                if math.isnan(col[1]):
                    col[1] = 0.0
                if math.isnan(col[2]):
                    col[2] = 0.0
                coord = Vector(((float(col[0])*0.001), (float(col[2])*0.001), (float(col[1]))* -0.001))
                if len(order_of_markers) > 0:
                    empty = order_of_markers[current_marker]
                    empty.location = coord
                    current_marker += 1 
                    for index in range(len(virtual_markers)):
                        update_virtual_marker(index)
            
            
            #keyframe bones
            #Goes through each bone
            for editBone in get_armature().data.edit_bones:
                boneName = editBone.name
                poseBone = arm.pose.bones[boneName]
                poseBone.keyframe_insert('rotation_euler', frame=scene.frame_current)
                poseBone.keyframe_insert('location', frame=scene.frame_current)
                poseBone.keyframe_insert('scale', frame=scene.frame_current)

        #-----------------------------------------------------------------------------------
        #script to create a mesh of the armature 
        def CreateMesh():
            obj = get_armature()

            if obj == None:
                print( "No selection" )
            elif obj.type != 'ARMATURE':
                print( "Armature expected" )
            else:
                return processArmature( bpy.context, obj )

        #Use armature to create base object
        def armToMesh( arm ):
            name = arm.name + "_mesh"
            dataMesh = bpy.data.meshes.new( name + "Data" )
            mesh = bpy.data.objects.new( name, dataMesh )
            mesh.matrix_world = arm.matrix_world.copy()
            return mesh

        #Make vertices and faces 
        def boneGeometry( l1, l2, x, z, baseSize, l1Size, l2Size, base ):
            x1 = x  * baseSize * l1Size * 0.8
            z1 = z  * baseSize * l2Size * 0.8
            
            x2 = Vector( (0, 0, 0) )
            z2 = Vector( (0, 0, 0) )

            verts = [
                l1 - x1 + z1,
                l1 + x1 + z1,
                l1 - x1 - z1,
                l1 + x1 - z1,
                l2 - x2 + z2,
                l2 + x2 + z2,
                l2 - x2 - z2,
                l2 + x2 - z2
                ] 

            faces = [
                (base+3, base+1, base+0, base+2),
                (base+6, base+4, base+5, base+7),
                (base+4, base+0, base+1, base+5),
                (base+7, base+3, base+2, base+6),
                (base+5, base+1, base+3, base+7),
                (base+6, base+2, base+0, base+4)
                ]

            return verts, faces

        #Process the armature, goes through its bones and creates the mesh
        def processArmature(context, arm, genVertexGroups = True):
            print("processing armature {0}".format(arm.name))

            #Creates the mesh object
            meshObj = armToMesh( arm )
            context.collection.objects.link( meshObj )

            verts = []
            edges = []
            faces = []
            vertexGroups = {}

            bpy.ops.object.mode_set(mode='EDIT')

            try:
                #Goes through each bone
                for editBone in get_armature().data.edit_bones:
                    boneName = editBone.name
                    if boneName == "":
                        editBone.name = "bone"
                        boneName = "bone"
                    poseBone = arm.pose.bones[boneName]

                    #Gets edit bone informations
                    editBoneHead = editBone.head
                    editBoneTail = editBone.tail
                    editBoneVector = editBoneTail - editBoneHead
                    editBoneSize = editBoneVector.dot( editBoneVector )
                    editBoneRoll = editBone.roll
                    editBoneX = editBone.x_axis
                    editBoneZ = editBone.z_axis
                    editBoneHeadRadius = editBone.head_radius
                    editBoneTailRadius = editBone.tail_radius

                    #Creates the mesh data for the bone
                    baseIndex = len(verts)
                    baseSize = math.sqrt( editBoneSize )
                    newVerts, newFaces = boneGeometry( editBoneHead, editBoneTail, editBoneX, editBoneZ, baseSize, editBoneHeadRadius, editBoneTailRadius, baseIndex )
                    verts.extend( newVerts )
                    faces.extend( newFaces )

                    #Creates the weights for the vertex groups
                    vertexGroups[boneName] = [(x, 1.0) for x in range(baseIndex, len(verts))]
                #Assigns the geometry to the mesh
                meshObj.data.from_pydata(verts, edges, faces)

            except:
                bpy.ops.object.mode_set(mode='OBJECT')
            else:
                bpy.ops.object.mode_set(mode='OBJECT')
            #Assigns the vertex groups
            if genVertexGroups:
                for name1, vertexGroup in vertexGroups.items():
                    groupObject = meshObj.vertex_groups.new(name = name1)
                    for (index, weight) in vertexGroup:
                        groupObject.add([index], weight, 'REPLACE')

            #Creates the armature modifier
            modifier = meshObj.modifiers.new('ArmatureMod', 'ARMATURE')
            modifier.object = arm
            modifier.use_bone_envelopes = False
            modifier.use_vertex_groups = True

            meshObj.data.update()

            return meshObj

        mesh_obob = CreateMesh()

        #-----------------------------------------------------------------------------------
        # Clean up the mesh by removing duplicate vertices, make sure all faces are quads, etc

        checked = set()
        for selected_object in bpy.data.objects:
            if selected_object.type != 'MESH':
                continue
            meshdata = selected_object.data
            if meshdata in checked:
                continue
            else:
                checked.add(meshdata)
            bpy.context.view_layer.objects.active = selected_object
            bpy.ops.object.editmode_toggle()
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.remove_doubles()
            bpy.ops.mesh.tris_convert_to_quads()
            bpy.ops.mesh.normals_make_consistent()
            bpy.ops.object.editmode_toggle()
        #Set armature active
        bpy.context.view_layer.objects.active = armature_data
        #Set armature selected
        armature_data.select_set(state=True)
        #-----------------------------------------------------------------------------------
        #function to register custom handler
        bpy.app.handlers.frame_change_post.clear()
        def register():
            bpy.app.handlers.frame_change_post.append(my_handler)
        
        def unregister():
            bpy.app.handlers.frame_change_post.remove(my_handler)
            
        register()

        scene = bpy.context.scene

        bpy.ops.object.mode_set(mode='OBJECT')

        #set keyframes for whole animation
        for frame in range(scene.frame_start, scene.frame_end):
            scene.frame_set(frame)
        
        #Select objects to export
        col = bpy.data.collections.get("Collection")
        if col:
            for obj in col.objects:
                if "Armature" == obj.name:
                        obj.select_set(True)

        
        ##########
        #### end Copy for fmc_blender ^^^
        ##########
        
        return{'FINISHED'}        




import bpy

bpy.ops.mesh.primitive_monkey_add(enter_editmode=False, align='WORLD', location=(0, 0, 0), scale=(1, 1, 1))

this_monke = bpy.context.active_object

bpy.ops.mesh.primitive_monkey_add(enter_editmode=False, align='WORLD', location=(0, 2, 0), scale=(1, 1, 1))

